import unittest

from parameterized import parameterized

from covigator import SYNONYMOUS_VARIANT
from covigator.precomputations.loader import PrecomputationsLoader
from covigator.precomputations.load_ns_s_counts import NsSCountsLoader
from covigator.database.model import JobStatus, DataSource, Gene, RegionType, Domain
from covigator.database.queries import Queries
from covigator.tests.unit_tests.abstract_test import AbstractTest
from covigator.tests.unit_tests.mocked import get_mocked_variant, \
    get_mocked_variant_observation, mock_samples, mock_cooccurrence_matrix, mock_samples_and_variants, MOCKED_DOMAINS, \
    get_mocked_sample


class QueriesTests(AbstractTest):

    def setUp(self) -> None:
        self.queries = Queries(session=self.session)

    def test_get_date_of_first_ena_sample(self):
        first_sample_date = None
        # adds 50 loaded and 50 failed samples, the date should only take into account loaded samples
        samples = mock_samples(faker=self.faker, session=self.session, num_samples=50, source=DataSource.ENA.name) + \
                  mock_samples(faker=self.faker, session=self.session, job_status=JobStatus.FAILED_PROCESSING, num_samples=50,
                               source=DataSource.ENA.name)
        for sample_ena in samples:
            if sample_ena.status == JobStatus.FINISHED:
                if first_sample_date is None:
                    first_sample_date = sample_ena.collection_date
                if sample_ena.collection_date < first_sample_date:
                    first_sample_date = sample_ena.collection_date
        observed_date = self.queries.get_date_of_first_sample()
        self.assertEqual(observed_date, first_sample_date)

    def test_get_date_of_first_ena_sample_empty(self):
        observed_date = self.queries.get_date_of_first_sample()
        self.assertIsNone(observed_date)

    def test_get_date_of_most_recent_ena_sample(self):
        most_recent_sample_date = None
        # adds 50 loaded and 50 failed samples, the date should only take into account loaded samples
        samples = mock_samples(faker=self.faker, session=self.session, num_samples=50, source=DataSource.ENA.name) + \
                  mock_samples(faker=self.faker, session=self.session, job_status=JobStatus.FAILED_PROCESSING, num_samples=50,
                               source=DataSource.ENA.name)
        for sample_ena in samples:
            if sample_ena.status == JobStatus.FINISHED:
                if most_recent_sample_date is None:
                    most_recent_sample_date = sample_ena.collection_date
                if sample_ena.collection_date > most_recent_sample_date:
                    most_recent_sample_date = sample_ena.collection_date
        observed_date = self.queries.get_date_of_most_recent_sample()
        self.assertEqual(observed_date, most_recent_sample_date)

    def test_get_date_of_most_recent_ena_sample_empty(self):
        observed_date = self.queries.get_date_of_most_recent_sample()
        self.assertIsNone(observed_date)

    def test_get_cooccurrence_matrix_by_gene_no_data(self):
        PrecomputationsLoader(session=self.session).load_table_counts()
        data = self.queries.get_sparse_cooccurrence_matrix(gene_name="S", domain=None)
        self.assertEqual(data.shape, (0, 10))

    # TODO: make this test stable
    @unittest.skip
    def test_get_cooccurrence_matrix_by_gene(self):
        other_variants, variants = mock_cooccurrence_matrix(faker=self.faker, session=self.session)
        PrecomputationsLoader(session=self.session).load_table_counts()

        data = self.queries.get_variants_cooccurrence_matrix(gene_name="S", min_cooccurrence=1, test=True)
        self.assertIsNotNone(data)
        num_unique_variants = len(set([v.variant_id for v in variants if v.annotation != SYNONYMOUS_VARIANT]))
        self.assertEqual(data.shape[0], num_unique_variants * num_unique_variants)
        self.assertEqual(data.shape[1], 7)
        self.assertGreaterEqual(data[data["count"] > 0].shape[0], len(variants))
        self.assertGreaterEqual(data[data["frequency"] > 0].shape[0], len(variants))
        self.assertGreaterEqual(data[data["jaccard"] > 0].shape[0], len(variants))

        data = self.queries.get_variants_cooccurrence_matrix(gene_name="S", min_cooccurrence=11, test=True)
        self.assertIsNone(data)

        data = self.queries.get_variants_cooccurrence_matrix(gene_name="X", min_cooccurrence=1, test=True)
        self.assertIsNone(data)

        data = self.queries.get_variants_cooccurrence_matrix(gene_name="N", min_cooccurrence=1, test=True)
        self.assertIsNotNone(data)
        self.assertEqual(data.shape[1], 7)
        self.assertGreaterEqual(data[data["count"] > 0].shape[0], len(other_variants))
        self.assertGreaterEqual(data[data["frequency"] > 0].shape[0], len(variants))
        self.assertGreaterEqual(data[data["jaccard"] > 0].shape[0], len(variants))

    @unittest.skip
    def test_get_mds(self):
        mock_cooccurrence_matrix(faker=self.faker, session=self.session)

        mds_fit, mds_coords = self.queries.get_mds(gene_name="S")
        self.assertIsNotNone(mds_fit)

    @parameterized.expand([(DataSource.ENA, ), (DataSource.GISAID, )])
    def test_get_variant_abundance_histogram(self, source):

        # gets an empty histogram
        histogram = self.queries.get_variant_abundance_histogram(cache=False, source=source.name)
        self.assertIsNone(histogram)

        # mocks a 100 variants
        num_variants = 100
        variants = [get_mocked_variant(faker=self.faker, chromosome="chr_test", source=source.name) for _ in range(num_variants)]
        self.session.add_all(variants)
        self.session.commit()

        # gets an histogram over 100 variants without variant observations
        histogram = self.queries.get_variant_abundance_histogram(cache=False, source=source.name)
        self.assertIsNotNone(histogram)
        self.assertGreater(histogram.shape[0], 0)
        self.assertEqual(histogram.shape[1], 3)
        self.assertEqual(histogram.count_unique_variants.sum(), num_variants)
        self.assertEqual(histogram.count_variant_observations.sum(), 0)

        # mock some variant observations
        test_samples = [get_mocked_sample(faker=self.faker, source=source) for _ in range(5)]
        self.session.add_all(test_samples)
        self.session.commit()
        variant_observations = []
        for v in variants:
            for i in range(self.faker.random_int(min=1, max=5)):
                variant_observations.append(get_mocked_variant_observation(variant=v, sample=test_samples[i]))
        self.session.add_all(variant_observations)
        self.session.commit()

        # gets an histogram over 100 variants
        histogram = self.queries.get_variant_abundance_histogram(cache=False, source=source.name)
        self.assertIsNotNone(histogram)
        self.assertGreater(histogram.shape[0], 0)
        self.assertEqual(histogram.shape[1], 3)
        self.assertEqual(histogram.count_unique_variants.sum(), num_variants)
        self.assertEqual(histogram.count_variant_observations.sum(), len(variant_observations))

    def test_get_conservation_table(self):
        conservation50 = self.queries.get_conservation_table(bin_size=50)
        self.assertIsNotNone(conservation50)
        self.assertEqual(conservation50.shape[1], 4)
        self.assertGreater(conservation50.shape[0], 0)
        self.assertEqual(conservation50[conservation50.conservation.isna()].shape[0], 0)
        self.assertEqual(conservation50[conservation50.conservation_sarbecovirus.isna()].shape[0], 0)
        self.assertEqual(conservation50[conservation50.conservation_vertebrates.isna()].shape[0], 0)

        conservation10 = self.queries.get_conservation_table(bin_size=10)
        self.assertGreater(conservation10.shape[0], conservation50.shape[0])
        self.assertEqual(conservation10.shape[1], conservation50.shape[1])

        conservation5 = self.queries.get_conservation_table(bin_size=5)
        self.assertGreater(conservation5.shape[0], conservation50.shape[0])
        self.assertGreater(conservation5.shape[0], conservation10.shape[0])
        self.assertEqual(conservation5.shape[1], conservation50.shape[1])

        conservation5_smaller = self.queries.get_conservation_table(bin_size=5, start=10000, end=20000)
        self.assertGreater(conservation5.shape[0], conservation5_smaller.shape[0])
        self.assertEqual(conservation5.shape[1], conservation50.shape[1])

    def test_get_genes_metadata(self):
        genes = self.queries.get_genes()
        self.assertIsNotNone(genes)
        self.assertGreater(len(genes), 0)
        for g in genes:
            self.assertIsInstance(g, Gene)

    def test_get_genes(self):
        genes = self.queries.get_genes()
        self.assertIsNotNone(genes)
        self.assertGreater(len(genes), 0)
        for g in genes:
            self.assertIsInstance(g, Gene)
            self.assertIsNotNone(g.name)

    def test_get_gene(self):
        gene = self.queries.get_gene("S")
        self.assertIsNotNone(gene)
        self.assertIsInstance(gene, Gene)
        gene = self.queries.get_gene("NOEXISTO")
        self.assertIsNone(gene)

    def test_get_domains(self):
        domains = self.queries.get_domains()
        self.assertIsNotNone(domains)
        self.assertGreater(len(domains), 0)
        for d in domains:
            self.assertIsInstance(d, Domain)
            self.assertIsNotNone(d.name)
            self.assertIsNotNone(d.gene_name)

    def test_get_domains_by_gene(self):
        domains = self.queries.get_domains_by_gene("S")
        self.assertIsNotNone(domains)
        self.assertGreater(len(domains), 0)
        for d in domains:
            self.assertIsInstance(d, Domain)
            self.assertIsNotNone(d.name)
            self.assertEqual(d.gene_name, "S")

    def test_count_jobs_in_queue(self):
        mock_samples(faker=self.faker, session=self.session, job_status=JobStatus.QUEUED, num_samples=50)
        mock_samples(faker=self.faker, session=self.session, job_status=JobStatus.FINISHED, num_samples=50)

        count_jobs_in_queue_ena = self.queries.count_jobs_in_queue(DataSource.ENA)
        self.assertGreater(count_jobs_in_queue_ena, 0)
        count_jobs_in_queue_gisaid = self.queries.count_jobs_in_queue(DataSource.GISAID)
        self.assertGreater(count_jobs_in_queue_gisaid, 0)

        self.assertEqual(count_jobs_in_queue_ena + count_jobs_in_queue_gisaid, 50)

    def test_get_dnds_table(self):
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        nsSCountsLoader = NsSCountsLoader(session=self.session)
        nsSCountsLoader.load()

        data = self.queries.get_dnds_table(source=DataSource.ENA.name)
        self._assert_dnds_table(data)
        self.assertEqual(data[data.source != DataSource.ENA].shape[0], 0)      # no entries to other source

        data = self.queries.get_dnds_table(source=DataSource.GISAID.name)
        self._assert_dnds_table(data)
        self.assertEqual(data[data.source != DataSource.GISAID].shape[0], 0)  # no entries to other source

        countries = list(data.country.unique())[0:2]
        data = self.queries.get_dnds_table(source=DataSource.GISAID.name, countries=countries)
        self._assert_dnds_table(data)
        self.assertEqual(data[~data.country.isin(countries)].shape[0], 0)  # no entries to other country

        genes = ["S"]
        data = self.queries.get_dnds_table(source=DataSource.ENA.name, genes=genes)
        self._assert_dnds_table(data, has_domains=True)
        self.assertEqual(data[~data.region_name.isin(genes + MOCKED_DOMAINS)].shape[0], 0)  # no entries to other country

    def _assert_dnds_table(self, data, has_domains=False):
        self.assertIsNotNone(data)
        self.assertGreater(data.shape[0], 0)
        self.assertEqual(data.shape[1], 8)
        self.assertGreater(data[data.region_type == RegionType.GENE].shape[0], 0)
        if has_domains:
            self.assertGreater(data[data.region_type == RegionType.DOMAIN].shape[0], 0)
        self.assertEqual(data[(data.region_type != RegionType.GENE) & (data.region_type != RegionType.DOMAIN)].shape[0], 0)  # all entries to a gene or domain
        self.assertEqual(data[data.country.isna()].shape[0], 0)  # no empty countries
        self.assertEqual(data[data.month.isna()].shape[0], 0)  # no empty months
        self.assertEqual(data[data.region_name.isna()].shape[0], 0)  # no empty gene



from itertools import combinations
from sqlalchemy.orm import Session
from covigator.database.queries import Queries


class CooccurrenceMatrix:

    def compute(self, run_accession: str, source: str, session: Session, maximum_length: int = 10):

        assert run_accession is not None or run_accession == "", "Missing sample identifier"
        assert session is not None, "Missing DB session"

        queries = Queries(session=session)
        sample_id = run_accession

        # the order by position is important to ensure we store only half the matrix and the same half of the matrix
        variant_ids = queries.get_variant_ids_by_sample(sample_id, source=source, maximum_length=maximum_length)

        # process all pairwise combinations without repetitions including the diagoonal
        for (variant_id_one, variant_id_two) in list(combinations(variant_ids, 2)) + list(zip(variant_ids, variant_ids)):
            queries.increment_variant_cooccurrence(variant_id_one, variant_id_two, source)



from sqlalchemy.orm import Session
from covigator.database.model import DataSource, JobStatus
from covigator.database.queries import Queries
from covigator.pipeline.cooccurrence_matrix import CooccurrenceMatrix
from logzero import logger


class CooccurrenceMatrixLoader:

    def __init__(self, session: Session):
        self.session = session
        self.queries = Queries(session=self.session)
        self.cooccurrence_matrix = CooccurrenceMatrix()

    def load(self, data_source: str, maximum_length: int):

        # deletes the database before loading
        self.session.query(self.queries.get_variant_cooccurrence_klass(data_source)).delete()

        # iterates over every sample in FINISHED status and computes the cooccurrence matrix
        sample_klass = self.queries.get_sample_klass(data_source)
        count_samples = self.queries.count_samples(source=data_source, cache=False)
        computed = 0
        query = self.session.query(sample_klass).filter(sample_klass.status == JobStatus.FINISHED)
        for sample in self.queries.windowed_query(query=query, column=sample_klass.run_accession, windowsize=1000):
            self.cooccurrence_matrix.compute(sample.run_accession, data_source, self.session,
                                             maximum_length=maximum_length)
            computed += 1
            if computed % 1000 == 0:
                logger.info('Processed cooccurrence over {}/{} ({}) samples'.format(
                    computed, count_samples, round(float(computed) / count_samples, 3)))



import logging
import os
import logzero
from logzero import logger

import covigator
from covigator.exceptions import CovigatorDashBoardInitialisationError

DEFAULT_SUBCLONAL_AF_THR = 0.8
DEFAULT_LOW_COVERAGE_AF_THR = 0.05
DEFAULT_HORIZONTAL_COVERAGE_THR = 20
DEFAULT_MEAN_BQ_THR = 10
DEFAULT_MEAN_MQ_THR = 10
DEFAULT_MINIMUM_SEQUENCE_SIZE = 5980    # 20 % of the genome, 29903 bp


class Configuration:

    # file system storage
    ENV_COVIGATOR_STORAGE_FOLDER = "COVIGATOR_STORAGE_FOLDER"
    ENV_COVIGATOR_TEMP_FOLDER = "COVIGATOR_TEMP_FOLDER"
    ENV_COVIGATOR_DOWNLOAD_CONTENT_FOLDER = "COVIGATOR_DOWNLOAD_CONTENT_FOLDER"
    # database
    ENV_COVIGATOR_DB_PORT = "COVIGATOR_DB_PORT"
    ENV_COVIGATOR_DB_PASSWORD = "COVIGATOR_DB_PASSWORD"
    ENV_COVIGATOR_DB_USER = "COVIGATOR_DB_USER"
    ENV_COVIGATOR_DB_NAME = "COVIGATOR_DB_NAME"
    ENV_COVIGATOR_DB_HOST = "COVIGATOR_DB_HOST"
    ENV_COVIGATOR_DB_POOL_SIZE = "COVIGATOR_DB_POOL_SIZE"
    ENV_COVIGATOR_DB_MAX_OVERFLOW = "COVIGATOR_DB_MAX_OVERFLOW"
    ENV_COVIGATOR_TABLE_VERSION = "COVIGATOR_TABLE_VERSION"
    # dashboard
    ENV_COVIGATOR_DASHBOARD_HOST = "COVIGATOR_DASHBOARD_HOST"
    ENV_COVIGATOR_DASHBOARD_PORT = "COVIGATOR_DASHBOARD_PORT"
    # logs
    ENV_COVIGATOR_DASHBOARD_LOG_FILE = "COVIGATOR_DASHBOARD_LOG_FILE"
    ENV_COVIGATOR_PROCESSOR_LOG_FILE = "COVIGATOR_PROCESSOR_LOG_FILE"
    ENV_COVIGATOR_ACCESSOR_LOG_FILE = "COVIGATOR_ACCESSOR_LOG_FILE"
    # pipeline
    ENV_COVIGATOR_NEXTFLOW = "COVIGATOR_NEXTFLOW"
    ENV_COVIGATOR_WORKFLOW = "COVIGATOR_WORKFLOW"
    ENV_COVIGATOR_FORCE_PIPELINE = "COVIGATOR_FORCE_PIPELINE"
    ENV_COVIGATOR_SKIP_VCF_LOADING = "COVIGATOR_SKIP_VCF_LOADING"
    ENV_COVIGATOR_WORKFLOW_CPUS = "COVIGATOR_WORKFLOW_CPUS"
    ENV_COVIGATOR_WORKFLOW_MEMORY = "COVIGATOR_WORKFLOW_MEMORY"
    ENV_COVIGATOR_BATCH_SIZE = "COVIGATOR_BATCH_SIZE"
    ENV_COVIGATOR_RETRIES_DOWNLOAD = "COVIGATOR_RETRIES_DOWNLOAD"
    ENV_COVIGATOR_LOW_COVERAGE_THR = "COVIGATOR_LOW_COVERAGE_THR"
    ENV_COVIGATOR_SUBCLONAL_THR = "COVIGATOR_SUBCLONAL_THR"
    # references
    ENV_COVIGATOR_REF_FASTA = "COVIGATOR_REF_FASTA"
    # dask
    ENV_COVIGATOR_DASK_PORT = "COVIGATOR_DASK_PORT"
    # sample exclusion
    ENV_COVIGATOR_MEAN_MQ_THR = "COVIGATOR_MEAN_MQ_THR"
    ENV_COVIGATOR_MEAN_BQ_THR = "COVIGATOR_MEAN_BQ_THR"
    ENV_COVIGATOR_HORIZONTAL_COVERAGE_THR = "COVIGATOR_HORIZONTAL_COVERAGE_THR"
    ENV_COVIGATOR_MAX_SNVS = "COVIGATOR_MAX_SNVS"
    ENV_COVIGATOR_MAX_INSERTIONS = "COVIGATOR_MAX_INSERTIONS"
    ENV_COVIGATOR_MAX_DELETIONS = "COVIGATOR_MAX_DELETIONS"
    ENV_COVIGATOR_MIN_SEQUENCE_SIZE = "COVIGATOR_MIN_SEQUENCE_SIZE"

    def __init__(self, verbose=True):
        # local storage
        self.storage_folder = os.getenv(self.ENV_COVIGATOR_STORAGE_FOLDER, "/data/covigator")
        self.content_folder = os.getenv(self.ENV_COVIGATOR_DOWNLOAD_CONTENT_FOLDER)

        # database
        self.db_host = os.getenv(self.ENV_COVIGATOR_DB_HOST, "localhost")
        self.db_name = os.getenv(self.ENV_COVIGATOR_DB_NAME, "covigator")
        self.db_user = os.getenv(self.ENV_COVIGATOR_DB_USER, "covigator")
        self.db_password = os.getenv(self.ENV_COVIGATOR_DB_PASSWORD, "covigator")
        self.db_port = os.getenv(self.ENV_COVIGATOR_DB_PORT, "5432")
        self.db_pool_size = int(os.getenv(self.ENV_COVIGATOR_DB_POOL_SIZE, 5))
        self.db_max_overflow = int(os.getenv(self.ENV_COVIGATOR_DB_MAX_OVERFLOW, 10))
        self.db_table_version = os.environ.get(self.ENV_COVIGATOR_TABLE_VERSION, "_test")
        self.force_pipeline = os.environ.get(self.ENV_COVIGATOR_FORCE_PIPELINE, False)
        # this is useful for reloading metadata into the DB
        self.skip_vcf_loading = os.environ.get(self.ENV_COVIGATOR_SKIP_VCF_LOADING, False)

        # dashboard
        self.dash_host = os.getenv(self.ENV_COVIGATOR_DASHBOARD_HOST, "0.0.0.0")
        try:
            self.dash_port = int(os.getenv(self.ENV_COVIGATOR_DASHBOARD_PORT, "8050"))
        except ValueError as e:
            raise CovigatorDashBoardInitialisationError("The port needs to be a numeric value. " + str(e))

        # dask
        self.dask_port = os.getenv(self.ENV_COVIGATOR_DASK_PORT, "50218")

        # logs
        self.logfile_dash = os.getenv(self.ENV_COVIGATOR_DASHBOARD_LOG_FILE)
        self.logfile_accesor = os.getenv(self.ENV_COVIGATOR_ACCESSOR_LOG_FILE)
        self.logfile_processor = os.getenv(self.ENV_COVIGATOR_PROCESSOR_LOG_FILE)

        # references
        self.reference_genome = os.getenv(self.ENV_COVIGATOR_REF_FASTA)

        # pipeline
        self.nextflow = os.getenv(self.ENV_COVIGATOR_NEXTFLOW, "nextflow")
        self.workflow = os.getenv(self.ENV_COVIGATOR_WORKFLOW,
                                  "tron-bioinformatics/covigator-ngs-pipeline -r {version}".format(
                                      version=covigator.ANALYSIS_PIPELINE_VERSION))
        self.workflow_cpus = os.getenv(self.ENV_COVIGATOR_WORKFLOW_CPUS, "1")
        self.workflow_memory = os.getenv(self.ENV_COVIGATOR_WORKFLOW_MEMORY, "3g")
        self.batch_size = self.load_numeric_value(variable=self.ENV_COVIGATOR_BATCH_SIZE, default=1000)
        self.retries_download = self.load_numeric_value(variable=self.ENV_COVIGATOR_RETRIES_DOWNLOAD, default=3)
        self.low_coverage_threshold = self.load_float_value(
            variable=self.ENV_COVIGATOR_LOW_COVERAGE_THR, default=DEFAULT_LOW_COVERAGE_AF_THR)
        self.subclonal_threshold = self.load_float_value(
            variable=self.ENV_COVIGATOR_SUBCLONAL_THR, default=DEFAULT_SUBCLONAL_AF_THR)

        ## sample exclusion
        self.mean_mq_thr = self.load_numeric_value(variable=self.ENV_COVIGATOR_MEAN_MQ_THR, default=DEFAULT_MEAN_MQ_THR)
        self.mean_bq_thr = self.load_numeric_value(variable=self.ENV_COVIGATOR_MEAN_BQ_THR, default=DEFAULT_MEAN_BQ_THR)
        self.horizontal_coverage_thr = self.load_numeric_value(
            variable=self.ENV_COVIGATOR_HORIZONTAL_COVERAGE_THR, default=DEFAULT_HORIZONTAL_COVERAGE_THR)
        self.max_snvs = self.load_numeric_value(variable=self.ENV_COVIGATOR_MAX_SNVS, default=76)
        self.max_insertions = self.load_numeric_value(variable=self.ENV_COVIGATOR_MAX_INSERTIONS, default=10)
        self.max_deletions = self.load_numeric_value(variable=self.ENV_COVIGATOR_MAX_DELETIONS, default=10)
        self.min_sequence_size = self.load_numeric_value(variable=self.ENV_COVIGATOR_MIN_SEQUENCE_SIZE,
                                                         default=DEFAULT_MINIMUM_SEQUENCE_SIZE)

        # NOTE: the defaults are already set in the workflow config
        self.temp_folder = os.getenv(self.ENV_COVIGATOR_TEMP_FOLDER, "data/covigator-tmp")

        if verbose:
            self.log_configuration()

    def load_numeric_value(self, variable, default):
        try:
            value = int(os.getenv(variable, default))
        except ValueError as e:
            raise CovigatorDashBoardInitialisationError("{} needs to be a numeric value : {}".format(variable, str(e)))
        return value

    def load_float_value(self, variable, default):
        try:
            value = float(os.getenv(variable, default))
        except ValueError as e:
            raise CovigatorDashBoardInitialisationError("{} needs to be a float value : {}".format(variable, str(e)))
        return value

    def log_configuration(self):
        logger.info("Configuration")
        for k, v in self.__dict__.items():
            logger.info("{}={}".format(k, v))


def initialise_logs(logfile, sample_id: str = None):
    if logfile is not None:
        logzero.logfile(logfile)
    logzero.loglevel(logging.INFO)
    if sample_id is not None:
        logzero.formatter(logging.Formatter(
            '[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d ' + sample_id + '] %(message)s'))
    else:
        logzero.formatter(logging.Formatter(
            '[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d] %(message)s'))



from parameterized import parameterized
from dash_core_components import Markdown, Graph
from covigator.dashboard.figures.mutation_stats import MutationStatsFigures
from covigator.dashboard.figures.samples import SampleFigures
from covigator.dashboard.figures.recurrent_mutations import RecurrentMutationsFigures
from covigator.database.model import VariantType, DataSource
from covigator.exceptions import CovigatorQueryException
from covigator.precomputations.loader import PrecomputationsLoader
from covigator.precomputations.load_ns_s_counts import NsSCountsLoader
from covigator.database.queries import Queries
from covigator.tests.unit_tests.abstract_test import AbstractTest
from covigator.tests.unit_tests.mocked import mock_samples_and_variants, mock_samples


class FiguresTests(AbstractTest):

    def setUp(self) -> None:
        self.sample_figures = SampleFigures(queries=Queries(session=self.session))
        self.mutation_stats_figures = MutationStatsFigures(queries=Queries(session=self.session))
        self.variants_figures = RecurrentMutationsFigures(queries=Queries(session=self.session))

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_samples_by_country(self, source):
        # populates the ENA samples tables
        mock_samples(faker=self.faker, session=self.session, num_samples=100)
        if source is not None:
            figure = self.sample_figures.get_accumulated_samples_by_country_plot(
                min_samples=0, data_source=source)
            self.assertIsNotNone(figure)
            self.assertTrue(len(figure) == 2)
            self.assertIsInstance(figure[0], Graph)
            self.assertIsInstance(figure[1], Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.sample_figures.get_accumulated_samples_by_country_plot,
                data_source=source
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_samples_by_country_no_data(self, source):
        if source is not None:
            figure = self.sample_figures.get_accumulated_samples_by_country_plot(data_source=source)
            self.assertIsInstance(figure, Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.sample_figures.get_accumulated_samples_by_country_plot,
                data_source=source
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_variants_per_sample(self, source):
        # populates the ENA samples tables
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        PrecomputationsLoader(session=self.session).load_counts_variants_per_sample()
        if source is not None:
            figure = self.mutation_stats_figures.get_variants_per_sample_plot(data_source=source)
            self.assertIsNotNone(figure)
            self.assertTrue(len(figure) == 2)
            self.assertIsInstance(figure[0], Graph)
            self.assertIsInstance(figure[1], Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.mutation_stats_figures.get_variants_per_sample_plot,
                data_source=source
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_variants_per_sample_no_data(self, source):
        if source is not None:
            figure = self.mutation_stats_figures.get_variants_per_sample_plot(data_source=source)
            self.assertIsInstance(figure, Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.mutation_stats_figures.get_variants_per_sample_plot,
                data_source=source
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_substitutions(self, source):
        # populates the ENA samples tables
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        PrecomputationsLoader(session=self.session).load_count_substitutions()

        if source is not None:
            figure = self.mutation_stats_figures.get_substitutions_plot(
                variant_types=[VariantType.SNV.name], data_source=source)
            self.assertIsNotNone(figure)
            self.assertTrue(len(figure) == 2)
            self.assertIsInstance(figure[0], Graph)
            self.assertIsInstance(figure[1], Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.mutation_stats_figures.get_substitutions_plot,
                data_source=source, variant_types=[VariantType.SNV.name]
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_substitutions_no_data(self, source):
        if source is not None:
            figure = self.mutation_stats_figures.get_substitutions_plot(
                variant_types=[VariantType.SNV.name], data_source=source)
            self.assertIsInstance(figure, Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.mutation_stats_figures.get_substitutions_plot,
                data_source=source, variant_types=[VariantType.SNV.name]
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_needle_plot(self, source):
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        if source is not None:
            figure = self.variants_figures.get_variants_plot(
                gene_name="S", domain_name=None, selected_variants=None, bin_size=50, source=source)
            self.assertIsInstance(figure, Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.variants_figures.get_variants_plot,
                source=source, gene_name="S", bin_size=50, domain_name=None, selected_variants=None
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_needle_plot_no_data(self, source):
        if source is not None:
            figure = self.variants_figures.get_variants_plot(
                gene_name="S", domain_name=None, selected_variants=None, bin_size=50, source=source)
            self.assertIsInstance(figure, Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.variants_figures.get_variants_plot,
                source=source, gene_name="S", bin_size=50, domain_name=None, selected_variants=None
            )

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name, (None,)])
    def test_dn_ds_plot(self, source):
        if source is not None:
            figure = self.sample_figures.get_dnds_by_gene_plot(data_source=source)
            self.assertIsInstance(figure, Markdown)

        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        loader = NsSCountsLoader(session=self.session)
        loader.load()

        if source is not None:
            figure = self.sample_figures.get_dnds_by_gene_plot(data_source=source)
            self.assertIsInstance(figure, list)
            self.assertIsInstance(figure[0], Graph)
            self.assertIsInstance(figure[1], Markdown)
        else:
            self.assertRaises(
                CovigatorQueryException,
                self.sample_figures.get_dnds_by_gene_plot,
                data_source=source
            )



import numpy as np
from parameterized import parameterized

import covigator
import pkg_resources
from dask.distributed import Client
from covigator.database.model import Log, DataSource, CovigatorModule, JobStatus, SampleEna, SampleGisaid
from covigator.processor.ena_processor import EnaProcessor
from covigator.processor.gisaid_processor import GisaidProcessor
from covigator.tests.unit_tests.abstract_test import AbstractTest
from covigator.tests.unit_tests.faked_objects import FakeEnaProcessor, \
    FakeGisaidProcessor, FakeProcessorFailing
from covigator.tests.unit_tests.mocked import mock_samples


class ProcessorTests(AbstractTest):

    def setUp(self) -> None:

        dask_client = Client(n_workers=int(1), threads_per_worker=1)

        ena_processor = EnaProcessor(
            database=self.database, dask_client=dask_client, config=self.config, wait_time=1)
        gisaid_processor = GisaidProcessor(
            database=self.database, dask_client=dask_client, config=self.config, wait_time=1)

        self.processors = {
            DataSource.ENA: ena_processor,
            DataSource.GISAID: gisaid_processor
        }

        fake_ena_processor = FakeEnaProcessor(
            database=self.database, dask_client=dask_client, config=self.config)
        fake_gisaid_processor = FakeGisaidProcessor(
            database=self.database, dask_client=dask_client, config=self.config)
        self.fake_processors = {
            DataSource.ENA: fake_ena_processor,
            DataSource.GISAID: fake_gisaid_processor
        }

        fake_ena_processor_fail = FakeProcessorFailing(
            database=self.database, dask_client=dask_client, config=self.config, source=DataSource.ENA)
        fake_gisaid_processor_fail = FakeProcessorFailing(
            database=self.database, dask_client=dask_client, config=self.config, source=DataSource.GISAID)
        self.failing_processors = {
            DataSource.ENA: fake_ena_processor_fail,
            DataSource.GISAID: fake_gisaid_processor_fail
        }

    @parameterized.expand([(DataSource.ENA, ), (DataSource.GISAID, )])
    def test_no_jobs(self, source):
        self.processors.get(source).process()
        self.assertEqual(self.session.query(Log).count(), 1)
        log = self.session.query(Log).first()
        self.assertIsNotNone(log.start)
        self.assertIsNotNone(log.end)
        self.assertEqual(log.source, source)
        self.assertEqual(log.module, CovigatorModule.PROCESSOR)
        self.assertEqual(log.processed, 0)
        data = log.data
        self.assertEqual(data.get("processed"), 0)

    @parameterized.expand([(DataSource.ENA, ), (DataSource.GISAID, )])
    def test_fake_processor(self, source):
        mock_samples(faker=self.faker, session=self.session, num_samples=10, job_status=JobStatus.DOWNLOADED,
                     source=source.name)

        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.DOWNLOADED), 10)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.FINISHED), 0)
        self.fake_processors.get(source).process()
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.DOWNLOADED), 0)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.FINISHED), 10)
        finished_jobs = self.queries.find_first_by_status(data_source=source, status=[JobStatus.FINISHED], n=10)
        for j in finished_jobs:
            self.assertEqual(j.status, JobStatus.FINISHED)
            self.assertIsNotNone(j.analysed_at)
            self.assertIsNotNone(j.pangolin_lineage)

    # @parameterized.expand([(DataSource.ENA,), (DataSource.GISAID,)])
    def test_processor(self, source=DataSource.ENA):
        samples = mock_samples(faker=self.faker, session=self.session, num_samples=1, job_status=JobStatus.DOWNLOADED,
                     source=source.name)
        sample = samples[0]

        fastq1 = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test_data_1.fastq.gz")
        fastq2 = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test_data_2.fastq.gz")
        sample.fastq_path = "{fastq1},{fastq2}".format(fastq1=fastq1, fastq2=fastq2)
        self.session.merge(sample)

        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.DOWNLOADED), 1)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.FINISHED), 0)
        self.fake_processors.get(source).process()
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.DOWNLOADED), 0)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.FINISHED), 1)

    @parameterized.expand([(DataSource.ENA, ), (DataSource.GISAID, )])
    def test_does_not_process_already_queued(self, source):
        mock_samples(faker=self.faker, session=self.session, num_samples=10, job_status=JobStatus.QUEUED,
                     source=source.name)

        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.PENDING), 0)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.QUEUED), 10)
        self.fake_processors.get(source).process()
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.PENDING), 0)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.QUEUED), 10)

    @parameterized.expand([(DataSource.ENA, ), (DataSource.GISAID, )])
    def test_failed_sample(self, source):
        mock_samples(faker=self.faker, session=self.session, num_samples=10, job_status=JobStatus.DOWNLOADED,
                     source=source.name)

        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.DOWNLOADED), 10)
        self.failing_processors.get(source).process()
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.DOWNLOADED), 0)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.FINISHED), 0)
        self.assertEqual(self.queries.count_jobs_by_status(data_source=source, status=JobStatus.FAILED_PROCESSING), 10)

    def test_load_pangolin_from_ena(self):
        pangolin_path = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test.lofreq.pangolin.csv")
        sample = SampleEna(run_accession="TEST", fastq_ftp="", fastq_md5="", num_fastqs=1)
        EnaProcessor.load_pangolin(sample, path=pangolin_path)
        self.assertEqual(sample.pangolin_pangolin_version, "3.1.19")
        self.assertEqual(sample.pangolin_version, "PLEARN-v1.2.123")
        self.assertEqual(sample.pangolin_pangoLEARN_version, "2022-01-20")
        self.assertEqual(sample.pangolin_pango_version, "v1.2.123")
        self.assertEqual(sample.pangolin_status, "passed_qc")
        self.assertEqual(sample.pangolin_note, "")
        self.assertEqual(sample.pangolin_lineage, "B.1.336")
        self.assertEqual(sample.pangolin_conflict, 0.0)
        self.assertEqual(sample.pangolin_ambiguity_score, 1.0)

        self.session.add(sample)
        self.session.commit()

    def test_load_pangolin_with_none(self):
        pangolin_path = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test.lofreq.pangolin.with_none.csv")
        sample = SampleEna(run_accession="TEST", fastq_ftp="", fastq_md5="", num_fastqs=1)
        EnaProcessor.load_pangolin(sample, path=pangolin_path)
        self.assertEqual(sample.pangolin_lineage, "")

        self.session.add(sample)
        self.session.commit()

    def test_load_pangolin_from_gisaid(self):
        pangolin_path = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test.lofreq.pangolin.csv")
        sample = SampleGisaid(run_accession="TEST")
        GisaidProcessor.load_pangolin(sample, path=pangolin_path)
        self.assertEqual(sample.pangolin_pangolin_version, "3.1.19")
        self.assertEqual(sample.pangolin_version, "PLEARN-v1.2.123")
        self.assertEqual(sample.pangolin_pangoLEARN_version, "2022-01-20")
        self.assertEqual(sample.pangolin_pango_version, "v1.2.123")
        self.assertEqual(sample.pangolin_status, "passed_qc")
        self.assertEqual(sample.pangolin_note, "")
        self.assertEqual(sample.pangolin_lineage, "B.1.336")
        self.assertEqual(sample.pangolin_conflict, 0.0)
        self.assertEqual(sample.pangolin_ambiguity_score, 1.0)

        self.session.add(sample)
        self.session.commit()

    def test_load_dedup_metrics(self):
        deduplication_metrics_path = pkg_resources.resource_filename(
            covigator.tests.__name__, "resources/test.deduplication_metrics.txt")
        sample = SampleEna(run_accession="TEST", deduplication_metrics_path=deduplication_metrics_path, fastq_ftp="", fastq_md5="",
                           num_fastqs=1)
        EnaProcessor.load_deduplication_metrics(sample)
        self.assertEqual(sample.percent_duplication, 0.919339)
        self.assertEqual(sample.unpaired_reads_examined, 0)
        self.assertEqual(sample.read_pairs_examined, 352625)
        self.assertEqual(sample.secondary_or_supplementary_reads, 1972)
        self.assertEqual(sample.unmapped_reads, 0)
        self.assertEqual(sample.unpaired_read_duplicates, 0)
        self.assertEqual(sample.read_pair_duplicates, 324182)
        self.assertEqual(sample.read_pair_optical_duplicates, 0)

        self.session.add(sample)
        self.session.commit()

    def test_load_horizontal_coverage(self):
        horizontal_coverage_path = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test.coverage.tsv")
        sample = SampleEna(run_accession="TEST", horizontal_coverage_path=horizontal_coverage_path, fastq_ftp="",
                           fastq_md5="", num_fastqs=1)
        EnaProcessor.load_coverage_results(sample)
        self.assertEqual(sample.mean_depth, 8.19399)
        self.assertEqual(sample.mean_base_quality, 35.9)
        self.assertEqual(sample.mean_mapping_quality, 60.0)
        self.assertEqual(sample.num_reads, 1984)
        self.assertEqual(sample.covered_bases, 26057)
        self.assertEqual(sample.coverage, 87.1384)

        self.session.add(sample)
        self.session.commit()




import pandas as pd
import numpy as np
import dash_core_components as dcc
import dash_table
from logzero import logger
from sqlalchemy.orm import Session
from covigator.dashboard.figures.figures import Figures, MARGIN, TEMPLATE, PLOTLY_CONFIG, STYLES_STRIPPED, STYLE_HEADER, \
    STYLE_CELL
from covigator.database.model import SubclonalVariantObservation, VariantObservation, SampleEna
from covigator.exceptions import CovigatorQueryException
import plotly.express as px


class SubclonalVariantsQueries:

    def __init__(self, session: Session):
        self.session = session

    def get_top_occurring_subclonal_variants(self, top, gene_name, domain, min_vaf, order_by):

        if order_by == "score":
            order_by_clause = "order by score asc"
        elif order_by == "count":
            order_by_clause = "order by count_observations desc"
        elif order_by == "conservation":
            order_by_clause = "order by conservation asc"
        elif order_by == "vaf":
            order_by_clause = "order by median_vaf desc"
        else:
            raise CovigatorQueryException("Not supported order by of intrahost variants")

        # counts variants over those bins
        if domain is not None:
            where_gene = "pfam_name = '{domain}'".format(domain=domain)
        elif gene_name is not None:
            where_gene = "gene_name = '{gene_name}'".format(gene_name=gene_name)
        else:
            where_gene = "gene_name is not null"

        sql_query = """
        select variant_id, date, hgvs_p, gene_name, annotation_highest_impact, 
            cons_hmm_sars_cov_2, cons_hmm_sarbecovirus, cons_hmm_vertebrate_cov, pfam_name, pfam_description, vaf
            from {subclonal_variants_table_name} 
            where 
                vaf >= {min_vaf} and 
                variant_id in (
                    select t.variant_id from (
                        select 
                            count(*) as count_observations, 
                            max(cons_hmm_sars_cov_2) as conservation, 
                            (ln(count(*)) * max(cons_hmm_sars_cov_2))::float as score, 
                            percentile_disc(0.5) within group (order by vaf) as median_vaf,
                            variant_id
                        from {subclonal_variants_table_name} 
                        where 
                            annotation_highest_impact != 'synonymous_variant' and
                            vaf >= {min_vaf} and 
                            {where_gene} and
                            variant_id not in (
                                select distinct variant_id 
                                from {variants_table_name} 
                                where annotation_highest_impact != 'synonymous_variant' and gene_name is not null 
                            )
                        group by variant_id
                        {order_by}
                        limit {top}) as t);
            """.format(
            min_vaf=min_vaf,
            subclonal_variants_table_name=SubclonalVariantObservation.__tablename__,
            variants_table_name=VariantObservation.__tablename__,
            top=top,
            where_gene=where_gene,
            order_by=order_by_clause
        )
        return pd.read_sql_query(sql_query, self.session.bind)

    def get_count_samples_by_library_stratgey(self, variant_id, min_vaf):
        sql_query = """
        select count(*) as count_samples, library_strategy 
        from {samples_ena_table}
        where run_accession in (
            select sample 
            from {subclonal_variant_observations_table} 
            where variant_id='{variant_id}' and vaf >= {min_vaf}
            )
        group by library_strategy
        """.format(
            variant_id=variant_id,
            min_vaf=min_vaf,
            samples_ena_table=SampleEna.__tablename__,
            subclonal_variant_observations_table=SubclonalVariantObservation.__tablename__
        )
        return pd.read_sql_query(sql_query, self.session.bind)

    def get_count_samples_by_country(self, variant_id, min_vaf):
        sql_query = """
        select count(*) as count_samples, country, collection_date as date
        from {samples_ena_table}
        where run_accession in (
            select sample 
            from {subclonal_variant_observations_table} 
            where variant_id='{variant_id}' and vaf >= {min_vaf}
            )
        group by country, collection_date
        """.format(
            variant_id=variant_id,
            min_vaf=min_vaf,
            samples_ena_table=SampleEna.__tablename__,
            subclonal_variant_observations_table=SubclonalVariantObservation.__tablename__
        )
        data = pd.read_sql_query(sql_query, self.session.bind).astype(
                {'date': 'datetime64', 'count_samples': 'float64'})

        data['cumsum'] = data.sort_values("date").groupby(['country'])['count_samples'].cumsum()
        dates = data.date[~data.date.isna()].sort_values().unique()
        countries = data[(~data.country.isna())].sort_values("cumsum", ascending=False).country.unique()

        empty_table = pd.DataFrame(
            index=pd.MultiIndex.from_product([dates, countries], names=["date", "country"]))
        empty_table["count_samples"] = 0

        # adds values into empty table
        filled_table = pd.merge(
            left=empty_table.reset_index(),
            right=data.loc[:, ["date", "country", "count_samples"]],
            on=["date", "country"],
            how='left',
            suffixes=("_x", "_y")).fillna(0)
        filled_table["count_samples"] = filled_table.count_samples_x + filled_table.count_samples_y
        filled_table['cumsum'] = filled_table.groupby(['country'])['count_samples'].cumsum()

        return filled_table

    def get_top_cooccurring_clonal_variants(self, variant_id, min_vaf, gene_name, domain):
        if domain is not None:
            where_gene = "pfam_name = '{domain}'".format(domain=domain)
        elif gene_name is not None:
            where_gene = "gene_name = '{gene_name}'".format(gene_name=gene_name)
        else:
            where_gene = "gene_name is not null"

        sql_query = """
        select count(*) as count_samples, variant_id, hgvs_p, gene_name, annotation_highest_impact
        from {variant_observations_table}
        where sample in (
            select sample 
            from {subclonal_variant_observations_table} 
            where variant_id='{variant_id}' and vaf >= {min_vaf}
            )
        and {where_gene}
        group by variant_id, hgvs_p, gene_name, annotation_highest_impact
        order by count_samples desc
        limit 10
        """.format(
            variant_id=variant_id,
            min_vaf=min_vaf,
            variant_observations_table=VariantObservation.__tablename__,
            subclonal_variant_observations_table=SubclonalVariantObservation.__tablename__,
            where_gene=where_gene
        )
        return pd.read_sql_query(sql_query, self.session.bind)


class IntrahostMutationsFigures(Figures):

    def get_top_occurring_subclonal_variants_plot(self, top, gene_name, domain, min_vaf, order_by):

        logger.debug("Getting data on top occurring intrahost mutations...")
        data = SubclonalVariantsQueries(session=self.queries.session).get_top_occurring_subclonal_variants(
            top=top, gene_name=gene_name, domain=domain, min_vaf=min_vaf, order_by=order_by)
        fig = dcc.Markdown("""**No intrahost variants for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Preparing plot on top occurring intrahost variants...")
            unique_subclonal_variants = data.groupby('variant_id').agg({
                'gene_name': 'first',
                'pfam_description': 'first',
                'hgvs_p': 'first',
                'annotation_highest_impact': 'first',
                'cons_hmm_sars_cov_2': 'first',
            })
            # this fills empty conservation values
            unique_subclonal_variants["cons_hmm_sars_cov_2"].fillna(0.0, inplace=True)
            unique_subclonal_variants["cons_hmm_sars_cov_2"] = unique_subclonal_variants["cons_hmm_sars_cov_2"].transform(lambda x: round(x, 3))
            unique_subclonal_variants["median_vaf"] = data[["variant_id", "vaf"]]\
                .groupby('variant_id')["vaf"].agg(lambda x: round(float(np.median(x)), 3))
            unique_subclonal_variants["iqr_vaf"] = data[["variant_id", "vaf"]]\
                .groupby('variant_id')["vaf"].agg(lambda x: round(np.percentile(x, 0.75) - np.percentile(x, 0.25),  3))
            unique_subclonal_variants["first_observation"] = data[["variant_id", "date"]] \
                .groupby('variant_id')["date"].agg(np.min)
            unique_subclonal_variants["last_observation"] = data[["variant_id", "date"]] \
                .groupby('variant_id')["date"].agg(np.max)
            unique_subclonal_variants["count_observations"] = data[["variant_id", "date"]] \
                .groupby('variant_id')["date"].agg('count')
            unique_subclonal_variants["score"] = unique_subclonal_variants[["count_observations", "cons_hmm_sars_cov_2"]]\
                .apply(lambda x: round(np.log(x[0]) * x[1], 3), axis=1)

            unique_subclonal_variants.reset_index(inplace=True)

            if order_by == "score":
                ordered_data = unique_subclonal_variants.sort_values("score", ascending=True)
            elif order_by == "count":
                ordered_data = unique_subclonal_variants.sort_values("count_observations", ascending=False)
            elif order_by == "conservation":
                ordered_data = unique_subclonal_variants.sort_values("cons_hmm_sars_cov_2", ascending=True)
            elif order_by == "vaf":
                ordered_data = unique_subclonal_variants.sort_values("median_vaf", ascending=False)
            else:
                raise CovigatorQueryException("Not supported order by of intrahost variants")

            fig = dash_table.DataTable(
                id='top-occurring-subclonal-variants-table',
                data=ordered_data.to_dict('records'),
                columns=[
                            {"name": ["Gene"], "id": "gene_name"},
                            {"name": ["Pfam Domain"], "id": "pfam_description"},
                            {"name": ["DNA mutation"], "id": "variant_id"},
                            {"name": ["Protein mutation"], "id": "hgvs_p"},
                            {"name": ["Effect"], "id": "annotation_highest_impact"},
                            {"name": ["First observation"], "id": "first_observation"},
                            {"name": ["Last observation"], "id": "last_observation"},
                            {"name": ["Count"], "id": "count_observations"},
                            {"name": ["ConsHMM"], "id": "cons_hmm_sars_cov_2"},
                            {"name": ["Median VAF"], "id": "median_vaf"},
                            {"name": ["IQR VAF"], "id": "iqr_vaf"},
                            {"name": ["Score"], "id": "score"},
                        ],
                style_data_conditional=STYLES_STRIPPED,
                style_as_list_view=True,
                style_header=STYLE_HEADER,
                row_selectable='single',
                css=[{'selector': '.dash-cell div.dash-cell-value',
                      'rule': 'display: inline; white-space: inherit; overflow: inherit; text-overflow: inherit;'}],
                style_table={'overflowX': 'auto'},
                style_data={'whiteSpace': 'normal', 'height': 'auto'},
                style_cell={'maxWidth': '100px'},
            )

        return [
            fig,
            dcc.Markdown("""
            **Mutations only observed as intrahost**
                            
            The list of intrahost mutations can be prioritised by the count of observations, 
            by the ConsHMM conservation score, by the median VAF or by a joint score using the count of observations 
            and the conservation scores (ie: *ln(count observations) x conservation score*).
            The variant calls can be filtered by VAF.
            
            Select any of the mutations to explore further details.
            """)
        ]

    def get_hist_library_strategy(self, variant_id, min_vaf):

        data = SubclonalVariantsQueries(session=self.queries.session).get_count_samples_by_library_stratgey(
            variant_id=variant_id, min_vaf=min_vaf)
        fig = px.bar(data_frame=data, x="library_strategy", y="count_samples", color_discrete_sequence=["#969696"])
        fig.update_layout(
            margin=MARGIN,
            template=TEMPLATE,
            yaxis={'title': "Num. of samples"},
            xaxis={'title': None},
            legend={'title': None}
        )
        return [
            dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
            dcc.Markdown("""**Library strategies distribution for {}**""".format(variant_id))
        ]

    def get_hist_countries(self, variant_id, min_vaf):

        data = SubclonalVariantsQueries(session=self.queries.session).get_count_samples_by_country(
            variant_id=variant_id, min_vaf=min_vaf)

        countries = list(data.sort_values("cumsum", ascending=False).country.unique())
        fig = px.area(data, x="date", y="cumsum", color="country",
                      category_orders={
                          "country": countries[::-1]},
                      labels={"cumsum": "num. samples", "count_samples": "increment"},
                      hover_data=["count_samples"],
                      color_discrete_sequence=px.colors.qualitative.Light24)

        fig.update_layout(
            margin=MARGIN,
            template=TEMPLATE,
            yaxis={'title': "Num. of samples"},
            xaxis={'title': None},
            legend={'title': None}
        )
        return [
            dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
            dcc.Markdown("""**Countries distribution for {}**""".format(variant_id))
        ]

    def get_cooccurring_clonal_variants(self, variant_id, min_vaf, gene_name, domain):
        data = SubclonalVariantsQueries(session=self.queries.session).get_top_cooccurring_clonal_variants(
            variant_id=variant_id, min_vaf=min_vaf, gene_name=gene_name, domain=domain)

        fig = dash_table.DataTable(
            id='top-cooccurring-clonal-variants-table',
            data=data.to_dict('records'),
            columns=[
                {"name": ["Gene"], "id": "gene_name"},
                {"name": ["DNA mutation"], "id": "variant_id"},
                {"name": ["Protein mutation"], "id": "hgvs_p"},
                {"name": ["Effect"], "id": "annotation_highest_impact"},
                {"name": ["Count samples"], "id": "count_samples"},
            ],
            style_data_conditional=STYLES_STRIPPED,
            style_as_list_view=True,
            style_header=STYLE_HEADER,
            css=[{'selector': '.dash-cell div.dash-cell-value',
                  'rule': 'display: inline; white-space: inherit; overflow: inherit; text-overflow: inherit;'}],
            style_table={'overflowX': 'auto'},
            style_data={'whiteSpace': 'normal', 'height': 'auto'},
            style_cell=STYLE_CELL,
        )

        return [
            fig,
            dcc.Markdown("""**Top 10 co-occurring clonal mutations with the intrahost mutation {variant_id}**
                    """.format(variant_id=variant_id))
        ]



from argparse import ArgumentParser

from covigator.precomputations.load_cooccurrences import CooccurrenceMatrixLoader
from covigator.precomputations.loader import PrecomputationsLoader
from dask.distributed import Client
from dask_jobqueue import SLURMCluster
import covigator
import covigator.configuration
from covigator.accessor.ena_accessor import EnaAccessor
from covigator.accessor.gisaid_accessor import GisaidAccessor
from covigator.configuration import Configuration
from covigator.database.database import Database
from covigator.pipeline.ena_pipeline import Pipeline
from covigator.pipeline.gisaid_pipeline import GisaidPipeline
from covigator.processor.ena_downloader import EnaDownloader
from covigator.processor.ena_processor import EnaProcessor
from covigator.processor.gisaid_processor import GisaidProcessor
from logzero import logger


def ena_accessor():
    parser = ArgumentParser(
        description="Covigator {} ENA accessor".format(covigator.VERSION))
    parser.add_argument(
        "--tax-id",
        dest="tax_id",
        help="the taxonomy id of the organism to analyse, eg: 2697049 for Sars-COV-2",
        default="2697049"
    )
    parser.add_argument(
        "--host-tax-id",
        dest="host_tax_id",
        help="the taxonomy id of the host organism, eg: 9606 for Homo sapiens",
        default="9606"
    )

    args = parser.parse_args()
    tax_id = args.tax_id
    host_tax_id = args.host_tax_id
    config = Configuration()
    covigator.configuration.initialise_logs(config.logfile_accesor)
    EnaAccessor(tax_id=tax_id, host_tax_id=host_tax_id, database=Database(config=config, initialize=True)).access()


def gisaid_accessor():
    parser = ArgumentParser(
        description="Covigator {} GISAID accessor".format(covigator.VERSION))
    parser.add_argument(
        "--input-fasta",
        dest="input_fasta",
        help="the FASTA input file from GISAID",
        required=True
    )
    parser.add_argument(
        "--input-metadata",
        dest="input_metadata",
        help="GISAID metadata file",
        required=True
    )

    args = parser.parse_args()

    config = Configuration()
    covigator.configuration.initialise_logs(config.logfile_accesor)
    GisaidAccessor(input_fasta=args.input_fasta, input_metadata=args.input_metadata, database=Database(initialize=True, config=config)).access()


def processor():
    parser = ArgumentParser(
        description="Covigator {} processor".format(covigator.VERSION))
    parser.add_argument(
        "--source",
        dest="data_source",
        help="Specify data source. This can be either ENA or GISAID",
        required=True
    )
    parser.add_argument(
        "--num-jobs",
        dest="num_jobs",
        help="The number of dask jobs to spin, this corresponds to the number of whole nodes requested to the cluster. "
             "The configuration of each job is in the correspoding jobqueue.yaml file",
        default=1
    )
    parser.add_argument(
        "--local",
        dest="local",
        help="if set it runs locally, no slurm cluster",
        action='store_true',
        default=False
    )
    parser.add_argument(
        "--num-local-cpus",
        dest="num_local_cpus",
        help="number of CPUs to be used by the processor when running locally",
        default=1
    )
    parser.add_argument(
        "--download",
        dest="download",
        help="if set it tries to download ENA FASTQs if they are not already in place. Not applicale for GISAID",
        action='store_true',
        default=False
    )

    args = parser.parse_args()
    config = Configuration()
    covigator.configuration.initialise_logs(config.logfile_processor)
    if args.local:
        _start_dask_processor(args, config, args.download, num_local_cpus=int(args.num_local_cpus))
    else:
        with SLURMCluster(
                walltime='72:00:00',  # hard codes maximum time to 72 hours
                scheduler_options={"dashboard_address": ':{}'.format(config.dask_port)}) as cluster:
            cluster.scale(jobs=int(args.num_jobs))
            _start_dask_processor(args, config, args.download, cluster=cluster)


def ena_downloader():
    parser = ArgumentParser(
        description="Covigator {} ENA downloader".format(covigator.VERSION))

    config = Configuration()
    covigator.configuration.initialise_logs(config.logfile_accesor)
    EnaDownloader(database=Database(config=config, initialize=True), config=config).process()


def _start_dask_processor(args, config, download, cluster=None, num_local_cpus=1):
    with Client(cluster) if cluster is not None else Client(n_workers=num_local_cpus, threads_per_worker=1) as client:
        if args.data_source == "ENA":
            EnaProcessor(database=Database(initialize=True, config=config),
                         dask_client=client, config=config, download=download) \
                .process()
        elif args.data_source == "GISAID":
            GisaidProcessor(database=Database(initialize=True, config=config),
                            dask_client=client, config=config) \
                .process()
        else:
            logger.error("Unknown data source. Please choose either ENA or GISAID")


def pipeline():
    parser = ArgumentParser(description="Run Pipeline for testing")
    parser.add_argument("--fastq1", dest="fastq1",
                        help="First FASTQ to process. In case of single end sequencing this will be the only FASTQ,"
                             "in the case of paired end sequencing this will be the first from the pair", required=True)
    parser.add_argument("--fastq2", dest="fastq2",
                        help="Second FASTQ to process for paired end sequencing, otherwise leave empty")

    args = parser.parse_args()
    vcf_file, qc_file = Pipeline(config=Configuration()).run(
        run_accession="test", fastq1=args.fastq1, fastq2=args.fastq2)
    logger.info("Output VCF file: {}".format(vcf_file))
    logger.info("Output QC file: {}".format(qc_file))


def gisaid_pipeline():
    parser = ArgumentParser(description="Run Pipeline for testing")
    parser.add_argument("--run_accession", dest="run_accession", help="Specify run accession to process", required=True)

    args = parser.parse_args()
    # FIXME: this interface is broken, it requires the SampleGisaid object not the run_accession
    vcf_file = GisaidPipeline(config=Configuration()).run(run_accession=args.run_accession)
    logger.info("Output VCF file: {}".format(vcf_file))


def precompute_queries():
    parser = ArgumentParser(description="Precompute some aggregation queries")
    parser.parse_args()

    database = Database(initialize=True, config=Configuration())
    loader = PrecomputationsLoader(session=database.get_database_session())
    logger.info("Starting precomputation...")
    loader.load()
    logger.info("Done precomputing")


def cooccurrence():
    parser = ArgumentParser(description="Precompute cooccurrence of mutations")
    parser.add_argument(
        "--source",
        dest="data_source",
        help="Specify data source. This can be either ENA or GISAID",
        required=True
    )
    parser.add_argument(
        "--maximum-mutation-length",
        dest="maximum_length",
        help="Only mutations with this maximum size will be included in the cooccurence matrix",
        default=10
    )
    args = parser.parse_args()

    database = Database(initialize=True, config=Configuration())
    loader = CooccurrenceMatrixLoader(session=database.get_database_session())
    logger.info("Starting precomputation...")
    loader.load(data_source=args.data_source, maximum_length=int(args.maximum_length))
    logger.info("Done precomputing")



from datetime import datetime
import dash_bootstrap_components as dbc

MONTH_PATTERN = "%Y-%m"
MISSING_VALUE = "-"


def print_date(date: datetime.date):
    return str(date) if date is not None else MISSING_VALUE


def print_number(value):
    return '{:,}'.format(value)


def get_mini_container(title, value):
    return dbc.Button(
        "{title}: {value}".format(title=title, value=value),
        outline=True,
        color="dark",
        className="me-1",
        style={"margin-right": "30px", "margin-left": "30px", 'font-size': '110%'})



import json
import numpy as np
from datetime import datetime

import pandas as pd

import covigator
from covigator.configuration import Configuration
from covigator.database.queries import Queries
from covigator.exceptions import CovigatorErrorProcessingCoverageResults, CovigatorExcludedSampleBadQualityReads, \
    CovigatorExcludedSampleNarrowCoverage, CovigatorErrorProcessingPangolinResults, \
    CovigatorErrorProcessingDeduplicationResults
from covigator.misc import backoff_retrier
from covigator.database.model import JobStatus, DataSource, SampleEna
from covigator.database.database import Database
from logzero import logger
from dask.distributed import Client
from covigator.processor.abstract_processor import AbstractProcessor
from covigator.pipeline.cooccurrence_matrix import CooccurrenceMatrix
from covigator.pipeline.downloader import Downloader
from covigator.pipeline.ena_pipeline import Pipeline
from covigator.pipeline.vcf_loader import VcfLoader


class EnaProcessor(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration, download : bool = False,
                 wait_time=60):
        logger.info("Initialising ENA processor")
        super().__init__(database, dask_client, DataSource.ENA, config, download=download, wait_time=wait_time)

    def _process_run(self, run_accession: str):
        """
        Launches all jobs and returns the futures for the final job only
        """
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(EnaProcessor.job, self.config, run_accession, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession):
        return EnaProcessor.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=DataSource.ENA,
            function=EnaProcessor.run_all)

    @staticmethod
    def run_all(sample: SampleEna, queries: Queries, config: Configuration) -> SampleEna:
        sample = EnaProcessor.download(sample=sample, queries=queries, config=config)
        sample = EnaProcessor.run_pipeline(sample=sample, queries=queries, config=config)
        sample = EnaProcessor.load(sample=sample, queries=queries, config=config)
        return sample

    @staticmethod
    def download(sample: SampleEna, queries: Queries, config: Configuration) -> SampleEna:
        # ensures that the download is done with retries, even after MD5 check sum failure
        downloader = Downloader(config=config)
        download_with_retries = backoff_retrier.wrapper(downloader.download, config.retries_download)
        paths = download_with_retries(sample_ena=sample)
        sample.sample_folder = sample.get_sample_folder(config.storage_folder)
        sample.fastq_path = paths
        sample.downloaded_at = datetime.now()
        return sample

    @staticmethod
    def run_pipeline(sample: SampleEna, queries: Queries, config: Configuration) -> SampleEna:
        fastq1, fastq2 = sample.get_fastq1_and_fastq2()
        pipeline_result = Pipeline(config=config)\
            .run(run_accession=sample.run_accession, fastq1=fastq1, fastq2=fastq2)
        sample.analysed_at = datetime.now()

        # stores the paths to all files output by pipeline
        sample.lofreq_vcf_path = pipeline_result.lofreq_vcf
        sample.ivar_vcf_path = pipeline_result.ivar_vcf
        sample.gatk_vcf_path = pipeline_result.gatk_vcf
        sample.bcftools_vcf_path = pipeline_result.bcftools_vcf
        sample.lofreq_pangolin_path = pipeline_result.lofreq_pangolin
        sample.ivar_pangolin_path = pipeline_result.ivar_pangolin
        sample.gatk_pangolin_path = pipeline_result.gatk_pangolin
        sample.bcftools_pangolin_path = pipeline_result.bcftools_pangolin
        sample.fastp_path = pipeline_result.fastp_qc
        sample.horizontal_coverage_path = pipeline_result.horizontal_coverage
        sample.vertical_coverage_path = pipeline_result.vertical_coverage
        sample.deduplication_metrics_path = pipeline_result.deduplication_metrics

        # stores the covigator version
        sample.covigator_processor_version = covigator.VERSION

        # load FAST JSON into the DB
        sample.qc = json.load(open(pipeline_result.fastp_qc))
        # load horizontal coverage values in the database
        sample = EnaProcessor.load_coverage_results(sample)
        # load deduplication metrics
        sample = EnaProcessor.load_deduplication_metrics(sample)
        # load pangolin results
        sample = EnaProcessor.load_pangolin(sample=sample, path=sample.lofreq_pangolin_path)

        return sample

    @staticmethod
    def load_deduplication_metrics(sample: SampleEna) -> SampleEna:
        try:
            data = pd.read_csv(sample.deduplication_metrics_path,
                               sep="\t",
                               skiprows=6,
                               nrows=1,
                               dtype={
                                   'PERCENT_DUPLICATION': float,
                                   'UNPAIRED_READS_EXAMINED': int,
                                   'READ_PAIRS_EXAMINED': int,
                                   'SECONDARY_OR_SUPPLEMENTARY_RDS': int,
                                   'UNMAPPED_READS': int,
                                   'UNPAIRED_READ_DUPLICATES': int,
                                   'READ_PAIR_DUPLICATES': int,
                                   'READ_PAIR_OPTICAL_DUPLICATES': int
                               })

            # fill NA values on a per column basis...
            data.PERCENT_DUPLICATION.fillna(value=0.0, inplace=True)
            data.UNPAIRED_READS_EXAMINED.fillna(value=0, inplace=True)
            data.READ_PAIRS_EXAMINED.fillna(value=0, inplace=True)
            data.SECONDARY_OR_SUPPLEMENTARY_RDS.fillna(value=0, inplace=True)
            data.UNMAPPED_READS.fillna(value=0, inplace=True)
            data.UNPAIRED_READ_DUPLICATES.fillna(value=0, inplace=True)
            data.READ_PAIR_DUPLICATES.fillna(value=0, inplace=True)
            data.READ_PAIR_OPTICAL_DUPLICATES.fillna(value=0, inplace=True)

            sample.percent_duplication = data.PERCENT_DUPLICATION.loc[0]
            # NOTE: SQLalchemy does not play well with int64
            sample.unpaired_reads_examined = int(data.UNPAIRED_READS_EXAMINED.loc[0])
            sample.read_pairs_examined = int(data.READ_PAIRS_EXAMINED.loc[0])
            sample.secondary_or_supplementary_reads = int(data.SECONDARY_OR_SUPPLEMENTARY_RDS.loc[0])
            sample.unmapped_reads = int(data.UNMAPPED_READS.loc[0])
            sample.unpaired_read_duplicates = int(data.UNPAIRED_READ_DUPLICATES.loc[0])
            sample.read_pair_duplicates = int(data.READ_PAIR_DUPLICATES.loc[0])
            sample.read_pair_optical_duplicates = int(data.READ_PAIR_OPTICAL_DUPLICATES.loc[0])
        except Exception as e:
            raise CovigatorErrorProcessingDeduplicationResults(e)

        return sample

    @staticmethod
    def load_coverage_results(sample: SampleEna) -> SampleEna:
        try:
            data = pd.read_csv(sample.horizontal_coverage_path,
                               sep="\t",
                               dtype={
                                   'numreads': int,
                                   'covbases': int,
                                   'meandepth': float,
                                   'meanbaseq': float,
                                   'meanmapq': float,
                                   'coverage': float,
                               }
                               )
            data.fillna(0, inplace=True)
            sample.mean_depth = float(data.meandepth.loc[0])
            sample.mean_base_quality = float(data.meanbaseq.loc[0])
            sample.mean_mapping_quality = float(data.meanmapq.loc[0])
            sample.num_reads = int(data.numreads.loc[0])
            sample.covered_bases = int(data.covbases.loc[0])
            sample.coverage = float(data.coverage.loc[0])
        except Exception as e:
            raise CovigatorErrorProcessingCoverageResults(e)

        return sample

    @staticmethod
    def load(sample: SampleEna, queries: Queries, config: Configuration) -> SampleEna:
        if sample.mean_mapping_quality < config.mean_mq_thr or sample.mean_base_quality < config.mean_bq_thr:
            raise CovigatorExcludedSampleBadQualityReads("Mean MQ: {}; mean BCQ: {}".format(
                sample.mean_mapping_quality, sample.mean_base_quality))
        if sample.coverage < config.horizontal_coverage_thr:
            raise CovigatorExcludedSampleNarrowCoverage("Horizontal coverage {} %".format(sample.coverage))
        if not config.skip_vcf_loading:
            VcfLoader().load(
                vcf_file=sample.lofreq_vcf_path, run_accession=sample.run_accession, source=DataSource.ENA, session=queries.session)
            sample.loaded_at = datetime.now()
        return sample



from itertools import cycle
import colorlover
import dash_table
import numpy as np
import re

import pandas as pd
from logzero import logger
from scipy.spatial.distance import squareform
from skbio.stats.distance import DissimilarityMatrix
from sklearn.cluster import OPTICS

from covigator import MISSENSE_VARIANT, DISRUPTIVE_INFRAME_DELETION, CONSERVATIVE_INFRAME_DELETION, \
    CONSERVATIVE_INFRAME_INSERTION, DISRUPTIVE_INFRAME_INSERTION
from covigator.dashboard.figures.figures import Figures, PLOTLY_CONFIG, TEMPLATE, MARGIN, STYLES_STRIPPED, STYLE_HEADER, \
    STYLE_CELL
import plotly
import plotly.express as px
import plotly.graph_objects as go
import dash_html_components as html
import dash_core_components as dcc
from covigator.database.model import Gene, Domain, DataSource

VARIANT_TOOLTIP = '<b>%{text}</b><br>' + 'Allele frequency: %{y:.5f}<br>' + 'Genomic Position: %{x}'
GENE_COLORS = list(reversed(plotly.express.colors.sequential.Tealgrn))
DOMAIN_COLORS = list(reversed(plotly.express.colors.sequential.Magenta))
OTHER_VARIANT_SYMBOL = "x"
INSERTION_SYMBOL = "triangle-up"
DELETION_SYMBOL = "triangle-down"
MISSENSE_VARIANT_SYMBOL = "circle"
VERY_COMMON_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-1]
COMMON_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-3]
RARE_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-7]
COMMON_VARIANTS_THRESHOLD = 0.1
LOW_FREQUENCY_VARIANTS_THRESHOLD = 0.01
LOW_FREQUENCY_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-5]
RARE_VARIANTS_THRESHOLD = 0.001
MONTH_PATTERN = re.compile('[0-9]{4}-[0-9]{2}')


class RecurrentMutationsFigures(Figures):

    def _get_color_by_af(self, af):
        color = None
        if af < RARE_VARIANTS_THRESHOLD:
            color = RARE_VARIANTS_COLOR
        elif RARE_VARIANTS_THRESHOLD <= af < LOW_FREQUENCY_VARIANTS_THRESHOLD:
            color = LOW_FREQUENCY_VARIANTS_COLOR
        elif LOW_FREQUENCY_VARIANTS_THRESHOLD <= af < COMMON_VARIANTS_THRESHOLD:
            color = COMMON_VARIANTS_COLOR
        elif af >= COMMON_VARIANTS_THRESHOLD:
            color = VERY_COMMON_VARIANTS_COLOR
        return color

    def _get_table_style_by_af(self):
        return [
            {
                'if': {
                    'filter_query': '{{frequency}} >= 0 && {{frequency}} < {}'.format(RARE_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': RARE_VARIANTS_COLOR,
                'color': 'inherit'
            },
            {
                'if': {
                    'filter_query': '{{frequency}} >= {} && {{frequency}} < {}'.format(
                        RARE_VARIANTS_THRESHOLD, LOW_FREQUENCY_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': LOW_FREQUENCY_VARIANTS_COLOR,
                'color': 'inherit'
            },
            {
                'if': {
                    'filter_query': '{{frequency}} >= {} && {{frequency}} < {}'.format(
                        LOW_FREQUENCY_VARIANTS_THRESHOLD, COMMON_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': COMMON_VARIANTS_COLOR,
                'color': 'white'
            },
            {
                'if': {
                    'filter_query': '{{frequency}} >= {}'.format(COMMON_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': VERY_COMMON_VARIANTS_COLOR,
                'color': 'white'
            }
        ]

    def get_top_occurring_variants_plot(self, top, gene_name, domain, date_range_start, date_range_end, metric, source):
        data = self.queries.get_top_occurring_variants_precomputed(top, gene_name, domain, metric, source)

        fig = [
            dash_table.DataTable(id="top-occurring-variants-table"),
            dcc.Markdown("""**No mutations for the current selection**""")
        ]

        if data is not None and data.shape[0] > 0:
            # removes the columns from the months out of the range
            month_columns = [c for c in data.columns if MONTH_PATTERN.match(c)]
            included_month_colums = [c for c in month_columns if c >= date_range_start and c <= date_range_end]
            excluded_month_colums = [c for c in month_columns if c < date_range_start or c > date_range_end]
            data.drop(excluded_month_colums, axis=1, inplace=True)

            # set the styles of the cells
            styles_counts = self.discrete_background_color_bins(data, columns=included_month_colums)
            styles_total_count = self.discrete_background_color_bins(data, columns=["total"], colors="Reds")
            styles_frequency = self._get_table_style_by_af()
            month_columns = [{'name': ["", i], 'id': i} for i in data.columns if i.startswith("20")]
            month_columns[0]['name'][0] = 'Monthly counts' if metric == "count" else 'Monthly frequencies'

            fig = [
                dash_table.DataTable(
                    # this cannot be imported from the tab definition due to a circular import...
                    id="top-occurring-variants-table",
                    data=data.to_dict('records'),
                    sort_action='native',
                    columns=[
                                {"name": ["Variant", "Gene"], "id": "gene_name"},
                                {"name": ["", "DNA mutation"], "id": "dna_mutation"},
                                {"name": ["", "Protein mutation"], "id": "hgvs_p"},
                                {"name": ["", "Effect"], "id": "annotation"},
                                {"name": ["", "Frequency"], "id": "frequency"},
                                {"name": ["", "Count"], "id": "total"},
                            ] + month_columns,
                    style_data_conditional=STYLES_STRIPPED + styles_counts + styles_frequency + styles_total_count,
                    style_table={'overflowX': 'auto'},
                    style_as_list_view=True,
                    style_header=STYLE_HEADER,
                    style_cell=STYLE_CELL,
                    sort_by=[{"column_id": "frequency", "direction": "desc"}],
                    row_selectable='multi'
                ),
                dcc.Markdown("""
                    **Top occurring mutations table** 
                    *table shows the {} mutations{} with the highest frequency across all samples.*
                    *The counts and frequencies per month are only shown between {} and {}.*
                    *Selections in this table will be highlighted in the genome view and in the co-occurrence matrix.*
                    """.format(top, " in gene {}".format(gene_name) if gene_name else "",
                               date_range_start, date_range_end))
            ]

        return fig

    def discrete_background_color_bins(self, df, n_bins=5, columns='all', colors='Blues'):
        bounds = [i * (1.0 / n_bins) for i in range(n_bins + 1)]
        if columns == 'all':
            if 'id' in df:
                df_numeric_columns = df.select_dtypes('number').drop(['id'], axis=1)
            else:
                df_numeric_columns = df.select_dtypes('number')
        else:
            df_numeric_columns = df[columns]
        df_max = df_numeric_columns.max().max()
        df_min = df_numeric_columns.min().min()
        ranges = [
            ((df_max - df_min) * i) + df_min
            for i in bounds
        ]
        styles = []
        for i in range(1, len(bounds)):
            min_bound = ranges[i - 1]
            max_bound = ranges[i]
            backgroundColor = colorlover.scales[str(n_bins)]['seq'][colors][i - 1]
            color = 'white' if i > len(bounds) / 2. else 'inherit'

            for column in df_numeric_columns:
                styles.append({
                    'if': {
                        'filter_query': (
                                '{{{column}}} >= {min_bound}' +
                                (' && {{{column}}} < {max_bound}' if (i < len(bounds) - 1) else '')
                        ).format(column=column, min_bound=min_bound, max_bound=max_bound),
                        'column_id': column
                    },
                    'backgroundColor': backgroundColor,
                    'color': color
                })

        return styles

    def get_cooccurrence_heatmap(self, sparse_matrix, selected_variants, metric="jaccard", min_cooccurrences=5):
        data = self._get_variants_cooccurrence_matrix(data=sparse_matrix)
        graph = dcc.Markdown("""**No co-occurrent mutations for the current selection**""")
        if data is not None and data.shape[0] > 0:

            all_variants = data.variant_id_one.unique()
            values = np.array_split(data[metric], len(all_variants))
            texts = np.array_split(data.hgvs_tooltip, len(all_variants))
            if metric == "count":
                hovertemplate = '<b>%{text}</b><br>' + 'Counts: %{z}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            elif metric == "frequency":
                hovertemplate = '<b>%{text}</b><br>' + 'Frequency: %{z:.3f}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            elif metric == "jaccard":
                hovertemplate = '<b>%{text}</b><br>' + 'Jaccard index: %{z:.3f}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            elif metric == "kappa":
                hovertemplate = '<b>%{text}</b><br>' + 'Kappa coefficient: %{z:.3f}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            heatmap = go.Heatmap(
                z=values,
                x=all_variants,
                y=all_variants,
                colorscale="Oryel",
                hoverongaps=False,
                text=texts,
                hovertemplate=hovertemplate,
            )
            if selected_variants:
                selected_variant_ids = [v.get("dna_mutation") for v in selected_variants]
                selected_data = data
                selected_data[metric].where(
                    (selected_data.variant_id_one.isin(selected_variant_ids)) |
                    (selected_data.variant_id_two.isin(selected_variant_ids)), np.nan, inplace=True)
                values_selected = np.array_split(selected_data[metric], len(all_variants))
                texts_selected = np.array_split(selected_data.hgvs_tooltip, len(all_variants))

                heatmap_selected = go.Heatmap(
                    z=values_selected,
                    x=all_variants,
                    y=all_variants,
                    colorscale="Teal",
                    hoverongaps=False,
                    showscale=False,
                    text=texts_selected,
                    hovertemplate=hovertemplate,
                )
            layout = go.Layout(
                template=TEMPLATE,
                margin=MARGIN,
                height=700,
                yaxis=dict(visible=True, tickfont={"size": 10}, showgrid=False, showspikes=True, spikemode='toaxis',
                           spikethickness=2),
                xaxis=dict(tickangle=-45, tickfont={"size": 10}, showgrid=False, showspikes=True, spikemode='toaxis',
                           spikethickness=2),
            )
            traces = [heatmap]
            if selected_variants:
                traces.append(heatmap_selected)
            fig = go.Figure(data=traces, layout=layout)

            # the y index is reversed in plotly heatmap
            fig.update_yaxes(autorange="reversed")
            graph = dcc.Graph(figure=fig, config=PLOTLY_CONFIG)

        return html.Div(children=[
            graph,
            dcc.Markdown("""
                        ***Co-occurrence matrix*** *showing variant pairs co-occurring in at least {} samples (this value is configurable).*
                        *The metric in the co-occurrence matrix can be chosen among counts, frequencies, Jaccard index or 
                        Cohen's kappa coefficient. The Cohen's kappa coefficient introduces a correction to the Jaccard index for
                        mutations with low occurrence.*
                        *The diagonal contains the total counts or just 1.0 in the other metrics.*
                        *The upper diagonal is not shown for clarity.*
                        *Synonymous mutations are excluded.*
                        *Different genomic mutations causing the same protein variant are not grouped.*
                        """.format(min_cooccurrences))
        ])

    def _get_variants_cooccurrence_matrix(self, data) -> pd.DataFrame:
        """
        Returns the full cooccurrence matrix of all non synonymous variants in a gene with at least
        min_occurrences occurrences.
        """
        # query for total samples required to calculate frequencies
        count_samples = self.queries.count_samples(source=DataSource.ENA.name)

        full_matrix = None
        if data.shape[0] > 0:
            # these are views of the original data
            annotations = data.loc[data.variant_id_one == data.variant_id_two,
                                   ["variant_id_one", "position", "reference", "alternate", "hgvs_p"]]
            tooltip = data.loc[:, ["variant_id_one", "variant_id_two", "hgvs_tooltip"]]
            diagonal = data.loc[
                data.variant_id_one == data.variant_id_two, ["variant_id_one", "variant_id_two", "count"]]
            sparse_matrix = data.loc[:, ["variant_id_one", "variant_id_two", "count"]]
            sparse_matrix["frequency"] = sparse_matrix["count"] / count_samples
            sparse_matrix = pd.merge(left=sparse_matrix, right=diagonal, on="variant_id_one", how="left",
                                     suffixes=("", "_one"))
            sparse_matrix = pd.merge(left=sparse_matrix, right=diagonal, on="variant_id_two", how="left",
                                     suffixes=("", "_two"))

            # calculate Jaccard index
            sparse_matrix["count_union"] = sparse_matrix["count_one"] + sparse_matrix["count_two"] - sparse_matrix[
                "count"]
            sparse_matrix["jaccard"] = sparse_matrix["count"] / sparse_matrix["count_union"]

            # calculate Cohen's kappa
            sparse_matrix["chance_agreement"] = np.exp(-sparse_matrix["count"])
            sparse_matrix["kappa"] = 1 - ((1 - sparse_matrix.jaccard) / (1 - sparse_matrix.chance_agreement))
            sparse_matrix["kappa"] = sparse_matrix["kappa"].transform(lambda k: k if k > 0 else 0)

            del sparse_matrix["count_union"]
            del sparse_matrix["count_one"]
            del sparse_matrix["count_two"]
            del sparse_matrix["chance_agreement"]

            # from the sparse matrix builds in memory the complete matrix
            all_variants = data.variant_id_one.unique()
            empty_full_matrix = pd.DataFrame(index=pd.MultiIndex.from_product(
                [all_variants, all_variants], names=["variant_id_one", "variant_id_two"])).reset_index()
            full_matrix = pd.merge(
                left=empty_full_matrix, right=sparse_matrix, on=["variant_id_one", "variant_id_two"], how='left')

            # add annotation on variant one
            full_matrix = pd.merge(left=full_matrix, right=annotations, on="variant_id_one", how='left') \
                .rename(columns={"position": "position_one",
                                 "reference": "reference_one",
                                 "alternate": "alternate_one",
                                 "hgvs_p": "hgvs_p_one"})

            # add annotations on variant two
            full_matrix = pd.merge(left=full_matrix, right=annotations,
                                   left_on="variant_id_two", right_on="variant_id_one", how='left') \
                .rename(columns={"variant_id_one_x": "variant_id_one",
                                 "position": "position_two",
                                 "reference": "reference_two",
                                 "alternate": "alternate_two",
                                 "hgvs_p": "hgvs_p_two"})

            # add tooltip
            full_matrix = pd.merge(left=full_matrix, right=tooltip, on=["variant_id_one", "variant_id_two"], how='left')
            # correct tooltip in diagonal
            full_matrix.loc[
                full_matrix.variant_id_one == full_matrix.variant_id_two, "hgvs_tooltip"] = full_matrix.hgvs_p_one

            # NOTE: transpose matrix manually as plotly transpose does not work with labels
            # the database return the upper diagonal, the lower is best for plots
            full_matrix.sort_values(["position_two", "reference_two", "alternate_two",
                                     "position_one", "reference_one", "alternate_one"], inplace=True)

            full_matrix = full_matrix.loc[:, ["variant_id_one", "variant_id_two", "count", "frequency", "jaccard",
                                              "kappa", "hgvs_tooltip"]]

        return full_matrix

    def get_variants_abundance_plot(self, source, bin_size=50):

        # reads genes and domains across the whole genome
        genes = self.queries.get_genes()
        domains = self.queries.get_domains()
        genes_and_domains = [(list(filter(lambda g: g.name == d.gene_name, genes))[0], d) for d in domains]

        # reads variants abundance
        variant_abundance = self.queries.get_variant_abundance_histogram(bin_size=bin_size, source=source)

        # reads conservation and bins it
        conservation = self.queries.get_conservation_table(bin_size=bin_size)

        # joins variant abundance and conservation
        data = variant_abundance.set_index("position_bin").join(conservation.set_index("position_bin"))
        data.reset_index(inplace=True)
        data.fillna(0, inplace=True)

        layout = go.Layout(
            template=TEMPLATE,
            margin=MARGIN,
            xaxis=dict(domain=[0, 1.0], tickformat=',d', hoverformat=',d', ticksuffix=" bp", ticks="outside",
                       visible=True, anchor="y7", showspikes=True, spikemode='across', spikethickness=2),
            yaxis=dict(domain=[0.9, 1.0], anchor='x7'),
            yaxis2=dict(domain=[0.6, 0.9], anchor='x7'),
            yaxis3=dict(domain=[0.45, 0.6], anchor='x7', visible=False),
            yaxis4=dict(domain=[0.3, 0.45], anchor='x7', visible=False),
            yaxis5=dict(domain=[0.15, 0.3], anchor='x7', visible=False),
            yaxis6=dict(domain=[0.05, 0.1], anchor='x7', visible=False),
            yaxis7=dict(domain=[0.0, 0.05], anchor='x7', visible=False),
            legend={'traceorder': 'normal'}
        )
        gene_colors = cycle(GENE_COLORS)
        domain_colors = cycle(DOMAIN_COLORS)
        gene_traces = [
            self._get_gene_trace(g, start=g.start, end=g.end, color=c, yaxis='y6') for g, c in zip(genes, gene_colors)]
        domain_traces = [self._get_domain_trace(color=c, domain=d, gene=g, yaxis='y7')
                         for (g, d), c in zip(genes_and_domains, domain_colors)]
        conservation_traces = self._get_conservation_traces(
            conservation, xaxis='x', yaxis1='y3', yaxis2='y4', yaxis3='y5')
        mean_unique_variants_per_bin = data.count_unique_variants.mean()
        variant_counts_traces = [
            go.Scatter(x=data.position_bin, y=data.count_variant_observations,
                       name="All variants", text="All variants", showlegend=False,
                       line_color=plotly.express.colors.sequential.Blues[-2], line_width=1),
            go.Scatter(x=data.position_bin,
                       y=[mean_unique_variants_per_bin for _ in range(data.shape[0])],
                       yaxis='y2', name="Mean unique variants", text="Mean unique variants",
                       line_width=1, showlegend=False, line_color=plotly.express.colors.sequential.Blues[-3]),
            go.Scatter(x=data.position_bin, y=data.count_unique_variants, yaxis='y2',
                       name="Unique variants", text="Unique variants", showlegend=False, fill='tonexty',
                       line_color=plotly.express.colors.sequential.Blues[-4], line_width=1)
        ]

        fig = go.Figure(data=variant_counts_traces + conservation_traces + gene_traces + domain_traces, layout=layout)

        # add track names
        fig.add_annotation(x=0.98, y=1.1, xref="x domain", yref="y domain", text="All variants",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=0.9, xref="x domain", yref="y2 domain", text="Unique variants",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y3 domain", text="Conservation SARS-CoV-2",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y4 domain", text="Conservation SARS-like betaCoV",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y5 domain", text="Conservation vertebrate CoV",
                           showarrow=False, yshift=10)

        return [
            dcc.Graph(
                figure=fig,
                config=PLOTLY_CONFIG
            ),
            dcc.Markdown("""
                ***Genome view*** *representing the abundance of mutations and ConsHMM (Arneson, 2019) conservation 
                using a bin size of {} bp. Synonymous mutations are included.*

                *The first track shows the count of mutations across the genome including repetitions. *
                *The second track shows the count of unique mutations across the genome, the horizontal line represents 
                the average number of unique mutations per bin and thus distinguishes regions with over and under the 
                average number of unique mutations.*
                *The third, fourth and fifth tracks represent the conservation as reported by ConsHMM within 
                SARS-CoV-2, among SARS-like betaCoV and among vertebrate CoV. Correlation between distribution of 
                unique mutations and conservation within Sars-CoV-2, among SARS-like betacoronavirus and among 
                vertebrates CoV respectively: {}, {}, {}.*
                *Genes and Pfam domains are represented in tones of red and purple respectively.*

                *Conservation data source: https://github.com/ernstlab/ConsHMM_CoV*

                *Arneson A, Ernst J. Systematic discovery of conservation states for single-nucleotide annotation of the 
                human genome. Communications Biology, 248, 2019. doi: https://doi.org/10.1038/s42003-019-0488-1*
                """.format(bin_size,
                           round(np.corrcoef(data.conservation, data.count_unique_variants)[0][1], 5),
                           round(np.corrcoef(data.conservation_sarbecovirus, data.count_unique_variants)[0][1], 5),
                           round(np.corrcoef(data.conservation_vertebrates, data.count_unique_variants)[0][1], 5)
                           ))]

    def get_variants_plot(self, gene_name, domain_name, selected_variants, bin_size, source):

        # reads gene annotations
        logger.debug("Getting genes and domains...")
        assert gene_name is not None or domain_name is not None, "Either gene or domain need to be provided"
        
        if domain_name is None:
            gene = self.queries.get_gene(gene_name)
            domains = self.queries.get_domains_by_gene(gene_name)
            start = gene.start
            end = gene.end
        else:
            domain = self.queries.get_domain(domain_name=domain_name)
            gene = self.queries.get_gene(domain.gene_name)
            domains = [domain]
            start = gene.start + (domain.start * 3)
            end = gene.start + (domain.end * 3)

        # reads variants
        logger.debug("Getting mutations...")
        variants = self.queries.get_non_synonymous_variants_by_region(start=start, end=end, source=source)

        # reads conservation and bins it
        logger.debug("Getting conservation...")
        conservation = self.queries.get_conservation_table(start=start, end=end, bin_size=bin_size)

        if variants.shape[0] > 0:
            # reads total number of samples and calculates frequencies
            logger.debug("Getting sample count...")
            count_samples = self.queries.count_samples(source=source)
            variants["af"] = variants.count_occurrences / count_samples
            variants["log_af"] = variants.af.transform(lambda x: np.log(x + 1))
            variants["log_count"] = variants.count_occurrences.transform(lambda x: np.log(x))
            variants.annotation_highest_impact = variants.annotation_highest_impact.transform(lambda a: a.split("&")[0])

            main_xaxis = 'x'

            variants_traces = []
            missense_variants = variants[variants.annotation_highest_impact == MISSENSE_VARIANT]
            if missense_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    missense_variants, name="missense variants", symbol=MISSENSE_VARIANT_SYMBOL, xaxis=main_xaxis))

            deletion_variants = variants[variants.annotation_highest_impact.isin(
                [DISRUPTIVE_INFRAME_DELETION, CONSERVATIVE_INFRAME_DELETION])]
            if deletion_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    deletion_variants, name="inframe deletions", symbol=DELETION_SYMBOL, xaxis=main_xaxis))

            insertion_variants = variants[variants.annotation_highest_impact.isin(
                [DISRUPTIVE_INFRAME_INSERTION, CONSERVATIVE_INFRAME_INSERTION])]
            if insertion_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    insertion_variants, name="inframe insertions", symbol=INSERTION_SYMBOL, xaxis=main_xaxis))

            other_variants = variants[~variants.annotation_highest_impact.isin([
                MISSENSE_VARIANT, DISRUPTIVE_INFRAME_DELETION, CONSERVATIVE_INFRAME_DELETION,
                DISRUPTIVE_INFRAME_INSERTION, CONSERVATIVE_INFRAME_DELETION])]
            if other_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    other_variants, name="other variants", symbol=OTHER_VARIANT_SYMBOL, xaxis=main_xaxis))

            selected_variants_trace = None
            if selected_variants:
                selected_variants_trace = go.Scatter(
                    x=[int(v.get("dna_mutation").split(":")[0]) for v in selected_variants],
                    y=[v.get("frequency") for v in selected_variants],
                    name="selected variants",
                    mode='markers',
                    marker=dict(
                        symbol="circle",
                        color="blue",
                        size=10,
                        showscale=False
                    ),
                    xaxis=main_xaxis,
                    showlegend=True,
                    text=["{} ({})".format(v.get("hgvs_p"), v.get("annotation")) for v in selected_variants],
                    hovertemplate=VARIANT_TOOLTIP
                )

            domain_colors = cycle(DOMAIN_COLORS)
            gene_trace = self._get_gene_trace(
                gene, start=start, end=end, color=plotly.express.colors.sequential.Tealgrn[-1], yaxis='y5', xaxis=main_xaxis)
            domain_traces = [self._get_domain_trace(
                color=c, gene=gene, domain=d, yaxis='y6', xaxis=main_xaxis, showlegend=True)
                for d, c in zip(domains, domain_colors)]
            conservation_traces = self._get_conservation_traces(
                conservation, main_xaxis, yaxis1='y2', yaxis2='y3', yaxis3='y4')

            data = variants_traces + conservation_traces + [gene_trace] + domain_traces
            if selected_variants_trace is not None:
                data.append(selected_variants_trace)

            layout = go.Layout(
                template=TEMPLATE,
                margin=go.layout.Margin(l=0, r=0, b=0, t=20),
                xaxis=dict(tickformat=',d', hoverformat=',d', ticksuffix=" bp", ticks="outside",
                           showspikes=True, spikemode='across', spikethickness=1, anchor='y6'),
                yaxis=dict(title='Allele frequency', type='log', domain=[0.4, 1.0], anchor=main_xaxis),
                yaxis2=dict(domain=[0.3, 0.4], visible=False, anchor=main_xaxis),
                yaxis3=dict(domain=[0.2, 0.3], visible=False, anchor=main_xaxis),
                yaxis4=dict(domain=[0.1, 0.2], visible=False, anchor=main_xaxis),
                yaxis5=dict(domain=[0.05, 0.1], visible=False, anchor=main_xaxis),
                yaxis6=dict(domain=[0.0, 0.05], visible=False, anchor=main_xaxis)
            )

            fig = go.Figure(data=data, layout=layout)

            fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y2 domain", text="Conservation SARS-CoV-2",
                               showarrow=False, yshift=10)
            fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y3 domain", text="Conservation SARS-like betaCoV",
                               showarrow=False, yshift=10)
            fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y4 domain", text="Conservation vertebrate CoV",
                               showarrow=False, yshift=10)

            # add horizontal lines on the frequency boundaries
            fig.add_hline(y=0.1, line_width=1, line_dash="dash", line_color=VERY_COMMON_VARIANTS_COLOR)
            fig.add_hline(y=0.01, line_width=1, line_dash="dash", line_color=COMMON_VARIANTS_COLOR)
            fig.add_hline(y=0.001, line_width=1, line_dash="dash", line_color=RARE_VARIANTS_COLOR)

            return [dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                    dcc.Markdown("""
                        ***Gene view*** *representing each variant with its frequency in the population and 
                        ConsHMM (Arneson, 2019) conservation using a bin size of {bin_size} bp. Synonymous mutations 
                        and mutations occurring in a single sample are excluded.*

                        *The scatter plot shows non synonymous mutations occurring in at least two samples on {region}. 
                        The x-axis shows the genomic coordinates and the y-axis shows the allele frequency.*
                        *The category of "other variants" includes frameshift indels, stop codon gain and lost and 
                        start lost mutations.*
                        *The mutations are colored according to their frequency as rare mutations (< 0.1 %), 
                        low frequency mutations (>= 0.1% and < 1%), common mutations (>= 1% and < 10%) and very common 
                        mutations (>= 10%). The second, third and fourth tracks in grey represent the conservation as 
                        reported by ConsHMM within SARS-CoV-2, among SARS-like betaCoV and among vertebrate CoV.*
                        *Genes and Pfam domains are represented in tones of red and purple respectively.*

                        *Conservation data source: https://github.com/ernstlab/ConsHMM_CoV*

                        *Arneson A, Ernst J. Systematic discovery of conservation states for single-nucleotide annotation of the 
                        human genome. Communications Biology, 248, 2019. doi: https://doi.org/10.1038/s42003-019-0488-1*
                        """.format(
                        bin_size=bin_size,
                        region="gene {}".format(gene_name) if domain_name is not None else
                                "domain {}: {}".format(gene_name, domain_name)))]
        else:
            return dcc.Markdown("""**No mutations for the current selection**""")

    def _get_conservation_traces(self, conservation, xaxis, yaxis1, yaxis2, yaxis3):
        return [
            go.Scatter(x=conservation.position_bin, y=conservation.conservation, yaxis=yaxis1, xaxis=xaxis,
                       text="Conservation SARS-CoV-2", textposition="top right", showlegend=False,
                       fill='tozeroy', line_color="grey", line_width=1),
            go.Scatter(x=conservation.position_bin, y=conservation.conservation_sarbecovirus, yaxis=yaxis2,
                       xaxis=xaxis, text="Conservation SARS-like betacoronavirus", textposition="top right",
                       showlegend=False, fill='tozeroy', line_color="grey", line_width=1),
            go.Scatter(x=conservation.position_bin, y=conservation.conservation_vertebrates, yaxis=yaxis3,
                       xaxis=xaxis, text="Conservation vertebrates", textposition="top right",
                       showlegend=False, fill='tozeroy', line_color="grey", line_width=1)
        ]

    @staticmethod
    def _get_domain_trace(color: str, domain: Domain, gene: Gene, yaxis='y', xaxis='x', showlegend=False):
        domain_start = gene.start + (domain.start * 3)  # domain coordinates are in the protein space
        domain_end = gene.start + (domain.end * 3)      # domain coordinates are in the protein space
        domain_name = domain.name
        return go.Scatter(
            mode='lines',
            x=[domain_start, domain_end, domain_end, domain_start],
            y=[0, 0, 1, 1],
            name=domain_name,
            fill="toself",
            fillcolor=color,
            text="<b>{} {}</b>: {}-{}".format(gene.name, domain_name, domain_start, domain_end),
            hoverinfo="text",
            line=dict(width=0),
            yaxis=yaxis,
            xaxis=xaxis,
            legendgroup='domains',
            showlegend=showlegend
        )

    @staticmethod
    def _get_gene_trace(gene: Gene, start, end, color, yaxis="y", xaxis='x'):
        # start and end coordinates in some case will come from the domain
        return go.Scatter(
            mode='lines',
            x=[start, end, end, start],
            y=[0, 0, 1, 1],
            name=gene.name,
            fill="toself",
            fillcolor=color,
            text="<b>{}</b>: {}-{}".format(gene.name, gene.start, gene.end),
            hoverinfo="text",
            line=dict(width=0),
            yaxis=yaxis,
            xaxis=xaxis,
            legendgroup='gene'
        )

    def _get_variants_scatter(self, variants, name, symbol, xaxis='x'):
        return go.Scatter(
            x=variants.position,
            y=variants.af,
            name=name,
            mode='markers',
            # opacity=0.5,
            marker=dict(
                symbol=symbol,
                color=variants.af.transform(self._get_color_by_af),
                showscale=False
            ),
            xaxis=xaxis,
            showlegend=True,
            text=variants[["hgvs_p", "annotation_highest_impact"]].apply(lambda x: "{} ({})".format(x[0], x[1]), axis=1),
            hovertemplate=VARIANT_TOOLTIP
        )

    def get_variants_clustering(self, sparse_matrix, min_cooccurrence, min_samples):

        data = self._get_mds(sparse_matrix=sparse_matrix, min_samples=min_samples)

        tables = []
        if data is not None:
            for c in data.cluster.unique():
                tables.append(dash_table.DataTable(
                    id="cluster{}-variants-table".format(c),
                    data=data[data.cluster == c].to_dict('records'),
                    columns=[
                        {"name": [
                            "Cluster {} (mean Jaccard={})".format(c, data[data.cluster == c].cluster_jaccard_mean.iloc[0]),
                            "DNA mutation"], "id": "variant_id"},
                        {"name": ["", "Protein mutation"], "id": "hgvs_p"},
                    ],
                    fixed_columns={'headers': True, 'data': 1},
                    style_table={'overflowX': 'auto'},
                    style_cell={'minWidth': '50px', 'width': '50px', 'maxWidth': '50px'},
                    style_as_list_view=True,
                    style_header={
                        'backgroundColor': 'rgb(230, 230, 230)',
                        'fontWeight': 'bold'
                    },
                    sort_by=[{"column_id": "variant_id", "direction": "asc"}],
                ))
                tables.append(html.Br())

        return html.Div(children=[
            html.Div(children=tables),
            dcc.Markdown("""
            ***Co-occurrence clustering*** *shows the resulting clusters from the
            co-occurrence matrix with the Jaccard index corrected with the Cohen's kappa coefficient. 
            The co-occurrence matrix is built taking into account only mutations with at least {} pairwise 
            co-occurrences and if a gene is provided only mutations within that gene.  
            Clustering is performed on the co-occurrence matrix using OPTICS. 
            The mimimum number of neighbours to call
            a cluster is {}.
            Variants selected in the top occurrent mutations table are highlighted with a greater size in the plot.
            *

            *Ankerst et al. “OPTICS: ordering points to identify the clustering structure.” ACM SIGMOD Record 28, no. 2 (1999): 49-60.*
            """.format(min_cooccurrence, min_samples))])

    def _get_mds(self, sparse_matrix, min_samples) -> pd.DataFrame:

        data = None

        if sparse_matrix is not None and sparse_matrix.shape[0] > 0:
            diagonal = sparse_matrix.loc[sparse_matrix.variant_id_one == sparse_matrix.variant_id_two,
                                         ["variant_id_one", "variant_id_two", "count"]]
            sparse_matrix_with_diagonal = pd.merge(
                left=sparse_matrix, right=diagonal, on="variant_id_one", how="left", suffixes=("", "_one"))
            sparse_matrix_with_diagonal = pd.merge(
                left=sparse_matrix_with_diagonal, right=diagonal, on="variant_id_two", how="left", suffixes=("", "_two"))

            # calculate Jaccard index
            sparse_matrix_with_diagonal["count_union"] = sparse_matrix_with_diagonal["count_one"] + \
                                                         sparse_matrix_with_diagonal["count_two"] - \
                                                         sparse_matrix_with_diagonal["count"]
            sparse_matrix_with_diagonal["jaccard_similarity"] = sparse_matrix_with_diagonal["count"] / \
                                                                sparse_matrix_with_diagonal.count_union
            sparse_matrix_with_diagonal["jaccard_dissimilarity"] = 1 - sparse_matrix_with_diagonal.jaccard_similarity

            # calculate Cohen's kappa
            sparse_matrix_with_diagonal["chance_agreement"] = np.exp(-sparse_matrix_with_diagonal["count"])
            sparse_matrix_with_diagonal["kappa"] = 1 - ((1 - sparse_matrix_with_diagonal.jaccard_similarity) / (
                    1 - sparse_matrix_with_diagonal.chance_agreement))
            sparse_matrix_with_diagonal["kappa"] = sparse_matrix_with_diagonal["kappa"].transform(
                lambda k: k if k > 0 else 0)
            sparse_matrix_with_diagonal["kappa_dissimilarity"] = 1 - sparse_matrix_with_diagonal.kappa

            dissimilarity_metric = "kappa_dissimilarity"

            # build upper diagonal matrix
            all_variants = sparse_matrix_with_diagonal.variant_id_one.unique()
            empty_full_matrix = pd.DataFrame(index=pd.MultiIndex.from_product(
                [all_variants, all_variants], names=["variant_id_one", "variant_id_two"])).reset_index()
            upper_diagonal_matrix = pd.merge(
                # gets only the inferior matrix without the diagnonal
                left=empty_full_matrix.loc[empty_full_matrix.variant_id_one < empty_full_matrix.variant_id_two, :],
                right=sparse_matrix_with_diagonal.loc[:, ["variant_id_one", "variant_id_two", dissimilarity_metric]],
                on=["variant_id_one", "variant_id_two"], how='left')
            upper_diagonal_matrix.fillna(1.0, inplace=True)
            upper_diagonal_matrix.sort_values(by=["variant_id_one", "variant_id_two"], inplace=True)

            logger.debug("Building square distance matrix...")
            distance_matrix = squareform(upper_diagonal_matrix[dissimilarity_metric])
            # this ensures the order of variants ids is coherent with the non redundant form of the distance matrix
            ids = np.array([list(upper_diagonal_matrix.variant_id_one[0])[0]] + \
                           list(upper_diagonal_matrix.variant_id_two[0:len(upper_diagonal_matrix.variant_id_two.unique())]))
            distance_matrix_with_ids = DissimilarityMatrix(data=distance_matrix, ids=ids)

            logger.debug("Clustering...")
            clusters = OPTICS(min_samples=min_samples, max_eps=1.4).fit_predict(distance_matrix_with_ids.data)

            logger.debug("Building clustering dataframe...")
            data = pd.DataFrame({'variant_id': distance_matrix_with_ids.ids, 'cluster': clusters})

            logger.debug("Annotate with HGVS.p ...")
            annotations = pd.concat([
                sparse_matrix.loc[:, ["variant_id_one", "hgvs_p_one"]].rename(
                    columns={"variant_id_one": "variant_id", "hgvs_p_one": "hgvs_p"}),
                sparse_matrix.loc[:, ["variant_id_two", "hgvs_p_two"]].rename(
                    columns={"variant_id_two": "variant_id", "hgvs_p_two": "hgvs_p"})])
            data = pd.merge(left=data, right=annotations, on="variant_id", how="left")

            logger.debug("Annotate with cluster mean Jaccard index...")
            data["cluster_jaccard_mean"] = 1.0
            for c in data.cluster.unique():
                variants_in_cluster = data[data.cluster == c].variant_id.unique()
                data.cluster_jaccard_mean = np.where(data.cluster == c, sparse_matrix_with_diagonal[
                    (sparse_matrix.variant_id_one.isin(variants_in_cluster)) &
                    (sparse_matrix.variant_id_two.isin(variants_in_cluster)) &
                    (sparse_matrix.variant_id_one != sparse_matrix.variant_id_two)
                    ].jaccard_dissimilarity.mean(), data.cluster_jaccard_mean)

            data.cluster_jaccard_mean = data.cluster_jaccard_mean.transform(lambda x: round(x, 3))

            data = data.set_index("variant_id").drop_duplicates().reset_index().sort_values("cluster")
            data = data[data.cluster != -1].loc[:, ["cluster", "variant_id", "hgvs_p", "cluster_jaccard_mean"]]

        return data



import covigator
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html


def get_tab_acknowledgements():

    return dbc.CardBody([
        dbc.Row([
            dbc.Col([None], width=2),
            dbc.Col([
                html.H1("Acknowledgements"),
                dcc.Markdown("""
                **The CoVigator project** was developed at the Biomarker Development Center at TRON (Translational 
                Oncology at the University Medical Center of the Johannes Gutenberg University) gGmbH. 
                The project was kindly supported by Intel´s Pandemic Response Technology Initiative.
                """),
                html.Br(),
                html.P("If you want to cite us:"),
                html.P([
                    "Schrörs, B., Riesgo-Ferreiro, P., Sorn, P., Gudimella, R., Bukur, T., Rösler, T., "
                    "Löwer, M., & Sahin, U. (2021). Large-scale analysis of SARS-CoV-2 spike-glycoprotein mutants "
                    "demonstrates the need for continuous screening of virus isolates. PLOS ONE, 16(9), e0249254.",
                ],
                    style={"font-style": "italic", "margin-left": "50px"}),
                html.P(
                    html.A("https://doi.org/10.1101/2021.02.04.429765",
                           href="https://doi.org/10.1101/2021.02.04.429765", target="_blank"),
                    style={"text-indent": "50px"}
                ),
                html.Br(),
                html.P("If you would like to cite the pipeline:"),
                html.P([
                    "Riesgo-Ferreiro, P., Sorn, P., & Bukur, T.. (2021, July 8). "
                    "TRON-Bioinformatics/covigator-ngs-pipeline. Zenodo. ",

                ], style={"font-style": "italic", "margin-left": "50px"}),
                html.P(
                    html.A("https://doi.org/10.5281/zenodo.4906281",
                           href="https://doi.org/10.5281/zenodo.4906281", target="_blank"),
                    style={"text-indent": "50px"}
                ),
                html.Br(),

                html.P(html.B("GISAID Data")),
                html.P([
                    """We gratefully acknowledge all data contributors, i.e. the Authors and their Originating laboratories 
                    responsible for obtaining the specimens, and their Submitting laboratories for generating the 
                    genetic sequence and metadata and sharing via the GISAID Initiative""",
                    html.Sup("1"),
                    " on which this research is based."]),
                html.P([
                    html.Sup("1"),
                    """ Elbe, S., and Buckland-Merrett, G. (2017) Data, disease and diplomacy: 
                    GISAID’s innovative contribution to global health. Global Challenges, 1:33-46. 
                    DOI: 10.1002/gch2.1018, PMCID: 31565258"""]),
                html.Br(),

                html.P(html.B("ENA Data")),
                html.P([
                    """We gratefully acknowledge the European Nucleotide Archive""",
                    html.Sup("2"),
                    " and all data contributors for sharing the raw reads on which this research is based."]),
                html.P([
                    html.Sup("2"),
                    """ Leinonen, R., Akhtar, R., Birney, E., Bower, L., Cerdeno-Tárraga, A., Cheng, Y., 
                    Cleland, I., Faruque, N., Goodgame, N., Gibson, R., Hoad, G., Jang, M., Pakseresht, N., 
                    Plaister, S., Radhakrishnan, R., Reddy, K., Sobhany, S., Hoopen, P. Ten, Vaughan, R., 
                    Zalunin V., Cochrane, G. (2011). The European nucleotide archive. Nucleic Acids Research, 
                    39(SUPPL. 1), D28. https://doi.org/10.1093/nar/gkq967"""]),
                html.Br(),

                html.P(html.B("About")),
                html.P("CoVigator dashboard version {}; analysis pipeline version {}".format(
                    covigator.VERSION, covigator.ANALYSIS_PIPELINE_VERSION
                )),
                html.P([
                    "Read the CoVigator documentation here ",
                    html.A("https://covigator.readthedocs.io",
                           href="https://covigator.readthedocs.io", target="_blank")
                ]),
                html.Br(),
                html.P("""
                       The CoVigator project is open sourced and made available under the MIT license.
                       We welcome any bug report, feature request or contribution through our GitHub repositories."""),
                html.P([
                    "The knowledge base and dashboard source is hosted at ",
                    html.A("https://github.com/TRON-Bioinformatics/covigator",
                           href="https://github.com/TRON-Bioinformatics/covigator", target="_blank")
                ]),
                html.P([
                    """
                    The analysis pipeline processes SARS-CoV-2 FASTQ or FASTA files into
                    annotated and normalized analysis ready VCF files. The pipeline is implemented
                    in the Nextflow framework (Di Tommaso, 2017), it is usable as an independent component
                    and we provide support to any user and to other viruses. The repository is hosted at
                    """,
                    html.A("https://github.com/TRON-Bioinformatics/covigator-ngs-pipeline",
                           href="https://github.com/TRON-Bioinformatics/covigator-ngs-pipeline", target="_blank")
                ]),

                html.Br(),
                html.P("We value your feedback!"),
                html.P("If you found an error or have a feature request, please, report it:"),
                html.Iframe(srcDoc="""
                        <a class="github-button" href="https://github.com/tron-bioinformatics/covigator/issues/new?assignees=&labels=bug&template=bug_report.md&title=" data-color-scheme="no-preference: light; light: light; dark: light;" data-size="large" data-show-count="false" aria-label="Issue tron-bioinformatics/covigator on GitHub">Issue</a>
                        <a class="github-button" href="https://github.com/tron-bioinformatics/covigator/issues/new?assignees=&labels=bug&template=feature_request.md&title=" data-color-scheme="no-preference: light; light: light; dark: light;" data-size="large" data-show-count="false" aria-label="Issue tron-bioinformatics/covigator on GitHub">Feature request</a>
                        <script async defer src="https://buttons.github.io/buttons.js"></script>
                       """, style={'border': 0, 'height': '50px'}),
                html.P("If you liked our work support us in GitHub:"),
                html.Iframe(srcDoc="""
                            <a class="github-button" href="https://github.com/tron-bioinformatics/covigator" data-color-scheme="no-preference: light; light: light; dark: light;" data-icon="octicon-star" data-size="large" data-show-count="true" aria-label="Star tron-bioinformatics/covigator on GitHub">Star</a>
                            <script async defer src="https://buttons.github.io/buttons.js"></script>
                           """, style={'border': 0, 'height': '50px'}),
                html.Br(),
                html.P("Follow #CoVigator in Twitter:"),
                html.A(html.Img(src="/assets/2021_twitter_logo_blue.png", height="25px"),
                       href="https://twitter.com/hashtag/CoVigator", target="_blank")

            ], width=8, style={"text-align": "left", "font-size": 16}),
            dbc.Col([None], width=2)]
            ),
    ])



from parameterized import parameterized
from sqlalchemy import and_, func

from covigator.database.model import PrecomputedSynonymousNonSynonymousCounts, RegionType, DataSource, \
    PrecomputedOccurrence, PrecomputedVariantsPerSample, PrecomputedSubstitutionsCounts, PrecomputedIndelLength, \
    PrecomputedAnnotation, PrecomputedVariantAbundanceHistogram, PrecomputedVariantsPerLineage, VariantCooccurrence
from covigator.precomputations.load_cooccurrences import CooccurrenceMatrixLoader
from covigator.precomputations.load_ns_s_counts import NsSCountsLoader
from covigator.precomputations.load_top_occurrences import TopOccurrencesLoader
from covigator.precomputations.load_variants_per_lineage import VariantsPerLineageLoader
from covigator.precomputations.loader import PrecomputationsLoader
from covigator.tests.unit_tests.abstract_test import AbstractTest
from covigator.tests.unit_tests.mocked import mock_samples_and_variants, MOCKED_GENES, MOCKED_DOMAINS


class TestPrecomputer(AbstractTest):

    def setUp(self) -> None:
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        self.ns_counts_loader = NsSCountsLoader(session=self.session)
        self.top_occurrences_loader = TopOccurrencesLoader(session=self.session)
        self.precomputations_loader = PrecomputationsLoader(session=self.session)
        self.precomputations_lineage = VariantsPerLineageLoader(session=self.session)

    def test_load_dn_ds(self):
        self.ns_counts_loader.load()
        for g in MOCKED_GENES:
            self.assertGreater(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.GENE.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == g)).count(),
                0)
            self.assertGreater(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.GENE.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == g,
                         PrecomputedSynonymousNonSynonymousCounts.source == DataSource.ENA.name)).count(),
                0)
            self.assertGreater(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.GENE.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == g,
                         PrecomputedSynonymousNonSynonymousCounts.source == DataSource.GISAID.name)).count(),
                0)
            self.assertEqual(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type != RegionType.GENE.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == g)).count(),
                0)
        self.assertGreater(
            self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.CODING_REGION.name)).count(), 0)
        self.assertGreater(
            self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.CODING_REGION.name,
                     PrecomputedSynonymousNonSynonymousCounts.source == DataSource.ENA.name)).count(), 0)
        self.assertGreater(
            self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.CODING_REGION.name,
                     PrecomputedSynonymousNonSynonymousCounts.source == DataSource.GISAID.name)).count(), 0)

        s_genes = self.session.query(func.sum(PrecomputedSynonymousNonSynonymousCounts.s)).filter(
            PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.GENE.name).scalar()
        s_coding_region = self.session.query(func.sum(PrecomputedSynonymousNonSynonymousCounts.s)).filter(
            PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.CODING_REGION.name).scalar()
        self.assertEqual(s_genes, s_coding_region)

        ns_genes = self.session.query(func.sum(PrecomputedSynonymousNonSynonymousCounts.ns)).filter(
            PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.GENE.name).scalar()
        ns_coding_region = self.session.query(func.sum(PrecomputedSynonymousNonSynonymousCounts.ns)).filter(
            PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.CODING_REGION.name).scalar()
        self.assertEqual(ns_genes, ns_coding_region)

        for d in MOCKED_DOMAINS:
            self.assertGreater(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.DOMAIN.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == d)).count(),
                0)
            self.assertGreater(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.DOMAIN.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == d,
                         PrecomputedSynonymousNonSynonymousCounts.source == DataSource.ENA.name)).count(),
                0)
            self.assertGreater(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.DOMAIN.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == d,
                         PrecomputedSynonymousNonSynonymousCounts.source == DataSource.GISAID.name)).count(),
                0)
            self.assertEqual(
                self.session.query(PrecomputedSynonymousNonSynonymousCounts).filter(
                    and_(PrecomputedSynonymousNonSynonymousCounts.region_type != RegionType.DOMAIN.name,
                         PrecomputedSynonymousNonSynonymousCounts.region_name == d)).count(),
                0)

    def test_load_table_counts(self):
        self.assertEqual(self.session.query(PrecomputedOccurrence).count(), 0)
        self.precomputations_loader.load_table_counts()     # table counts precomputations are needed
        self.top_occurrences_loader.load()
        self.assertGreater(self.session.query(PrecomputedOccurrence).count(), 0)
        for g in MOCKED_GENES:
            occurrences = self.session.query(PrecomputedOccurrence).filter(PrecomputedOccurrence.gene_name == g).all()
            self.assertGreater(len(occurrences), 0)
            for o in occurrences:
                self.assertGreater(o.total, 0)
                self.assertGreater(o.frequency, 0.0)
                self.assertIsNotNone(o.variant_id)
                self.assertIsNotNone(o.gene_name)
                self.assertIsNotNone(o.domain)
                self.assertIsNotNone(o.annotation)

    def test_load_counts_variants_per_sample(self):
        self.assertEqual(self.session.query(PrecomputedVariantsPerSample).count(), 0)
        self.precomputations_loader.load_counts_variants_per_sample()
        self.assertGreater(self.session.query(PrecomputedVariantsPerSample).count(), 0)
        p: PrecomputedVariantsPerSample
        for p in self.session.query(PrecomputedVariantsPerSample).all():
            self.assertGreater(p.count, 0)
            self.assertIsNotNone(p.source)
            self.assertIsNotNone(p.variant_type)
            self.assertIsNotNone(p.number_mutations)

    def test_load_count_substitutions(self):
        self.assertEqual(self.session.query(PrecomputedSubstitutionsCounts).count(), 0)
        self.precomputations_loader.load_count_substitutions()
        self.assertGreater(self.session.query(PrecomputedSubstitutionsCounts).count(), 0)
        p: PrecomputedSubstitutionsCounts
        for p in self.session.query(PrecomputedSubstitutionsCounts).all():
            self.assertGreater(p.count, 0)
            self.assertIsNotNone(p.source)
            self.assertIsNotNone(p.variant_type)
            self.assertIsNotNone(p.reference)
            self.assertIsNotNone(p.alternate)

    def test_load_indel_length(self):
        self.assertEqual(self.session.query(PrecomputedIndelLength).count(), 0)
        self.precomputations_loader.load_indel_length()
        # TODO: mock indels
        #self.assertGreater(self.session.query(PrecomputedIndelLength).count(), 0)
        #p: PrecomputedIndelLength
        #for p in self.session.query(PrecomputedIndelLength).all():
        #    self.assertGreater(p.count, 0)
        #    self.assertIsNotNone(p.source)
        #    self.assertIsNotNone(p.gene_name)
        #    self.assertIsNotNone(p.length)

    def test_load_annotation(self):
        self.assertEqual(self.session.query(PrecomputedAnnotation).count(), 0)
        self.precomputations_loader.load_annotation()
        self.assertGreater(self.session.query(PrecomputedAnnotation).count(), 0)
        p: PrecomputedAnnotation
        for p in self.session.query(PrecomputedAnnotation).all():
            self.assertGreater(p.count, 0)
            self.assertIsNotNone(p.source)
            self.assertIsNotNone(p.annotation)

    def test_load_variant_abundance_histogram(self):
        self.assertEqual(self.session.query(PrecomputedVariantAbundanceHistogram).count(), 0)
        self.precomputations_loader.load_variant_abundance_histogram()
        self.assertGreater(self.session.query(PrecomputedVariantAbundanceHistogram).count(), 0)
        p: PrecomputedVariantAbundanceHistogram
        for p in self.session.query(PrecomputedVariantAbundanceHistogram).all():
            self.assertGreaterEqual(p.count_variant_observations, 0)
            self.assertGreaterEqual(p.count_variant_observations, p.count_unique_variants)
            self.assertIsNotNone(p.source)
            self.assertIsNotNone(p.bin_size)
            self.assertIsNotNone(p.position_bin)

    def test_load_variants_per_lineage(self):
        self.assertEqual(self.session.query(PrecomputedVariantsPerLineage).count(), 0)
        self.precomputations_lineage.load()
        self.assertGreater(self.session.query(PrecomputedVariantsPerLineage).count(), 0)
        for p in self.session.query(PrecomputedVariantsPerLineage).all():
            self.assertGreater(p.count_observations, 0)
            self.assertIsNotNone(p.lineage)
            self.assertNotEqual(p.lineage, "")
            self.assertIsNotNone(p.variant_id)
            self.assertIsNotNone(p.country)

    @parameterized.expand([DataSource.ENA.name, DataSource.GISAID.name])
    def test_load_cooccurrence_matrix(self, source):

        variant_cooccurrence_klass = self.queries.get_variant_cooccurrence_klass(source)

        self.assertEqual(self.session.query(variant_cooccurrence_klass).count(), 0)
        CooccurrenceMatrixLoader(self.session).load(data_source=source, maximum_length=10)
        self.assertGreater(self.session.query(variant_cooccurrence_klass).count(), 0)
        for p in self.session.query(variant_cooccurrence_klass).all():
            self.assertGreater(p.count, 0)
            self.assertIsNotNone(p.variant_id_one)
            self.assertIsNotNone(p.variant_id_two)
